create table web_sales(
	ws_sold_date_sk integer,
	ws_sold_time_sk integer,
	ws_ship_date_sk integer,
	ws_item_sk integer not null,
	ws_bill_customer_sk integer,
	ws_bill_cdemo_sk integer,
	ws_bill_hdemo_sk integer,
	ws_bill_addr_sk integer,
	ws_ship_customer_sk integer,
	ws_ship_cdemo_sk integer,
	ws_ship_hdemo_sk integer,
	ws_ship_addr_sk integer,
	ws_web_page_sk integer,
	ws_web_site_sk integer,
	ws_ship_mode_sk integer,
	ws_warehouse_sk integer,
	ws_promo_sk integer,
	ws_order_number integer not null,
	ws_quantity integer,
	ws_wholesale_cost decimal(7,2),
	ws_list_price decimal(7,2),
	ws_sales_price decimal(7,2),
	ws_ext_discount_amt decimal(7,2),
	ws_ext_sales_price decimal(7,2),
	ws_ext_wholesale_cost decimal(7,2),
	ws_ext_list_price decimal(7,2),
	ws_ext_tax decimal(7,2),
	ws_coupon_amt decimal(7,2),
	ws_ext_ship_cost decimal(7,2),
	ws_net_paid decimal(7,2),
	ws_net_paid_inc_tax decimal(7,2),
	ws_net_paid_inc_ship decimal(7,2),
	ws_net_paid_inc_ship_tax decimal(7,2),
	ws_net_profit decimal(7,2)
);
