create table catalog_returns(
	cr_returned_date_sk integer,
	cr_returned_time_sk integer,
	cr_item_sk integer not null,
	cr_refunded_customer_sk integer,
	cr_refunded_cdemo_sk integer,
	cr_refunded_hdemo_sk integer,
	cr_refunded_addr_sk integer,
	cr_returning_customer_sk integer,
	cr_returning_cdemo_sk integer,
	cr_returning_hdemo_sk integer,
	cr_returning_addr_sk integer,
	cr_call_center_sk integer,
	cr_catalog_page_sk integer,
	cr_ship_mode_sk integer,
	cr_warehouse_sk integer,
	cr_reason_sk integer,
	cr_order_number integer not null,
	cr_return_quantity integer,
	cr_return_amount decimal(7,2),
	cr_return_tax decimal(7,2),
	cr_return_amt_inc_tax decimal(7,2),
	cr_fee decimal(7,2),
	cr_return_ship_cost decimal(7,2),
	cr_refunded_cash decimal(7,2),
	cr_reversed_charge decimal(7,2),
	cr_store_credit decimal(7,2),
	cr_net_loss decimal(7,2)
);
