create table store_sales(
	ss_sold_date_sk integer,
	ss_sold_time_sk integer,
	ss_item_sk integer not null,
	ss_customer_sk integer,
	ss_cdemo_sk integer,
	ss_hdemo_sk integer,
	ss_addr_sk integer,
	ss_store_sk integer,
	ss_promo_sk integer,
	ss_ticket_number integer not null,
	ss_quantity integer,
	ss_wholesale_cost decimal(7,2),
	ss_list_price decimal(7,2),
	ss_sales_price decimal(7,2),
	ss_ext_discount_amt decimal(7,2),
	ss_ext_sales_price decimal(7,2),
	ss_ext_wholesale_cost decimal(7,2),
	ss_ext_list_price decimal(7,2),
	ss_ext_tax decimal(7,2),
	ss_coupon_amt decimal(7,2),
	ss_net_paid decimal(7,2),
	ss_net_paid_inc_tax decimal(7,2),
	ss_net_profit decimal(7,2)
);