create table store(
	s_store_sk integer not null,
	s_store_id varchar(16) not null,
	s_rec_start_date date , s_rec_end_date date , s_closed_date_sk integer,
	s_store_name varchar(50),
	s_number_employees integer,
	s_floor_space integer,
	s_hours varchar(20),
	s_manager varchar(40),
	s_market_id integer,
	s_geography_class varchar(100),
	s_market_desc varchar(100),
	s_market_manager varchar(40),
	s_division_id integer,
	s_division_name varchar(50),
	s_company_id integer,
	s_company_name varchar(50),
	s_street_number varchar(10),
	s_street_name varchar(60),
	s_street_type varchar(15),
	s_suite_number varchar(10),
	s_city varchar(60),
	s_county varchar(30),
	s_state varchar(2),
	s_zip varchar(10),
	s_country varchar(20),
	s_gmt_offset decimal(5,2),
	s_tax_percentage decimal(5,2)
);
