#include "duckdb/storage/compression/chimp/algorithm/leading_zero_buffer.hpp"

namespace duckdb {

constexpr uint32_t LeadingZeroBufferConstants::MASKS[];
constexpr uint8_t LeadingZeroBufferConstants::SHIFTS[];

} // namespace duckdb
