create table date_dim(
	d_date_sk integer not null,
	d_date_id varchar(16) not null,
	d_date date , d_month_seq integer,
	d_week_seq integer,
	d_quarter_seq integer,
	d_year integer,
	d_dow integer,
	d_moy integer,
	d_dom integer,
	d_qoy integer,
	d_fy_year integer,
	d_fy_quarter_seq integer,
	d_fy_week_seq integer,
	d_day_name varchar(9),
	d_quarter_name varchar(6),
	d_holiday varchar(1),
	d_weekend varchar(1),
	d_following_holiday varchar(1),
	d_first_dom integer,
	d_last_dom integer,
	d_same_day_ly integer,
	d_same_day_lq integer,
	d_current_day varchar(1),
	d_current_week varchar(1),
	d_current_month varchar(1),
	d_current_quarter varchar(1),
	d_current_year varchar(1)
);
