#include "duckdb_python/import_cache/modules/pyarrow_module.hpp"
#include "duckdb_python/import_cache/modules/pandas_module.hpp"
#include "duckdb_python/import_cache/modules/datetime_module.hpp"
#include "duckdb_python/import_cache/modules/decimal_module.hpp"
#include "duckdb_python/import_cache/modules/ipython_module.hpp"
#include "duckdb_python/import_cache/modules/ipywidgets_module.hpp"
#include "duckdb_python/import_cache/modules/numpy_module.hpp"
#include "duckdb_python/import_cache/modules/pathlib_module.hpp"
#include "duckdb_python/import_cache/modules/polars_module.hpp"
#include "duckdb_python/import_cache/modules/duckdb_module.hpp"
#include "duckdb_python/import_cache/modules/pytz_module.hpp"
#include "duckdb_python/import_cache/modules/types_module.hpp"
#include "duckdb_python/import_cache/modules/typing_module.hpp"
#include "duckdb_python/import_cache/modules/uuid_module.hpp"
#include "duckdb_python/import_cache/modules/collections_module.hpp"