#ifndef ZIP_CODE_DATA_FILE_RECORD_H
#define ZIP_CODE_DATA_FILE_RECORD_H

/*
 * Legal Notice
 *
 * This document and associated source code (the "Work") is a part of a
 * benchmark specification maintained by the TPC.
 *
 * The TPC reserves all right, title, and interest to the Work as provided
 * under U.S. and international laws, including without limitation all patent
 * and trademark rights therein.
 *
 * No Warranty
 *
 * 1.1 TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THE INFORMATION
 *     CONTAINED HEREIN IS PROVIDED "AS IS" AND WITH ALL FAULTS, AND THE
 *     AUTHORS AND DEVELOPERS OF THE WORK HEREBY DISCLAIM ALL OTHER
 *     WARRANTIES AND CONDITIONS, EITHER EXPRESS, IMPLIED OR STATUTORY,
 *     INCLUDING, BUT NOT LIMITED TO, ANY (IF ANY) IMPLIED WARRANTIES,
 *     DUTIES OR CONDITIONS OF MERCHANTABILITY, OF FITNESS FOR A PARTICULAR
 *     PURPOSE, OF ACCURACY OR COMPLETENESS OF RESPONSES, OF RESULTS, OF
 *     WORKMANLIKE EFFORT, OF LACK OF VIRUSES, AND OF LACK OF NEGLIGENCE.
 *     ALSO, THERE IS NO WARRANTY OR CONDITION OF TITLE, QUIET ENJOYMENT,
 *     QUIET POSSESSION, CORRESPONDENCE TO DESCRIPTION OR NON-INFRINGEMENT
 *     WITH REGARD TO THE WORK.
 * 1.2 IN NO EVENT WILL ANY AUTHOR OR DEVELOPER OF THE WORK BE LIABLE TO
 *     ANY OTHER PARTY FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO THE
 *     COST OF PROCURING SUBSTITUTE GOODS OR SERVICES, LOST PROFITS, LOSS
 *     OF USE, LOSS OF DATA, OR ANY INCIDENTAL, CONSEQUENTIAL, DIRECT,
 *     INDIRECT, OR SPECIAL DAMAGES WHETHER UNDER CONTRACT, TORT, WARRANTY,
 *     OR OTHERWISE, ARISING IN ANY WAY OUT OF THIS OR ANY OTHER AGREEMENT
 *     RELATING TO THE WORK, WHETHER OR NOT SUCH AUTHOR OR DEVELOPER HAD
 *     ADVANCE NOTICE OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * Contributors
 * - Doug Johnson
 */

#include <deque>
#include <string>

namespace TPCE {
//
// Description:
// A class to represent a single record in the ZipCode data file.
//
// Exception Safety:
// The Basic guarantee is provided.
//
// Copy Behavior:
// Copying is allowed.
//

class ZipCodeDataFileRecord {
private:
	int divisionTaxKey;

	static const int maxZc_codeLen = 12;
	char zc_codeCStr[maxZc_codeLen + 1];
	std::string zc_code;

	static const int maxZc_townLen = 80;
	char zc_townCStr[maxZc_townLen + 1];
	std::string zc_town;

	static const int maxZc_divLen = 80;
	char zc_divCStr[maxZc_divLen + 1];
	std::string zc_div;

	static const unsigned int fieldCount = 4;

public:
	explicit ZipCodeDataFileRecord(const std::deque<std::string> &fields);

	//
	// Default copies and destructor are ok.
	//
	// ~ZipCodeDataFileRecord()
	// ZipCodeDataFileRecord(const ZipCodeDataFileRecord&);
	// ZipCodeDataFileRecord& operator=(const ZipCodeDataFileRecord&);
	//

	int DivisionTaxKey() const;

	const std::string &ZC_CODE() const;
	const char *ZC_CODE_CSTR() const;

	const std::string &ZC_TOWN() const;
	const char *ZC_TOWN_CSTR() const;

	const std::string &ZC_DIV() const;
	const char *ZC_DIV_CSTR() const;

	std::string ToString(char fieldSeparator = '\t') const;
};

} // namespace TPCE
#endif // ZIP_CODE_DATA_FILE_RECORD_H
