/*
 * THIS IS A GENERATED FILE
 * SEE COLUMNS.LIST
 */
#ifndef STREAMS_H
#define STREAMS_H
rng_t Streams[] = {
    {0, 0, 0, 0, 0, 0, 0},
    {0, 0, 0, 0, CC_CALL_CENTER_SK, CALL_CENTER, CC_CALL_CENTER_SK},
    {0, 15, 0, 0, CC_CALL_CENTER_ID, CALL_CENTER, CC_CALL_CENTER_ID},
    {0, 10, 0, 0, CC_REC_START_DATE_ID, CALL_CENTER, CC_REC_START_DATE_ID},
    {0, 1, 0, 0, CC_REC_END_DATE_ID, CALL_CENTER, CC_REC_END_DATE_ID},
    {0, 4, 0, 0, CC_CLOSED_DATE_ID, CALL_CENTER, CC_CLOSED_DATE_ID},
    {0, 10, 0, 0, CC_OPEN_DATE_ID, CALL_CENTER, CC_OPEN_DATE_ID},
    {0, 0, 0, 0, CC_NAME, CALL_CENTER, CC_NAME},
    {0, 2, 0, 0, CC_CLASS, CALL_CENTER, CC_CLASS},
    {0, 1, 0, 0, CC_EMPLOYEES, CALL_CENTER, CC_EMPLOYEES},
    {0, 1, 0, 0, CC_SQ_FT, CALL_CENTER, CC_SQ_FT},
    {0, 1, 0, 0, CC_HOURS, CALL_CENTER, CC_HOURS},
    {0, 2, 0, 0, CC_MANAGER, CALL_CENTER, CC_MANAGER},
    {0, 1, 0, 0, CC_MARKET_ID, CALL_CENTER, CC_MARKET_ID},
    {0, 50, 0, 0, CC_MARKET_CLASS, CALL_CENTER, CC_MARKET_CLASS},
    {0, 50, 0, 0, CC_MARKET_DESC, CALL_CENTER, CC_MARKET_DESC},
    {0, 2, 0, 0, CC_MARKET_MANAGER, CALL_CENTER, CC_MARKET_MANAGER},
    {0, 2, 0, 0, CC_DIVISION, CALL_CENTER, CC_DIVISION},
    {0, 2, 0, 0, CC_DIVISION_NAME, CALL_CENTER, CC_DIVISION_NAME},
    {0, 2, 0, 0, CC_COMPANY, CALL_CENTER, CC_COMPANY},
    {0, 2, 0, 0, CC_COMPANY_NAME, CALL_CENTER, CC_COMPANY_NAME},
    {0, 0, 0, 0, CC_STREET_NUMBER, CALL_CENTER, CC_STREET_NUMBER},
    {0, 0, 0, 0, CC_STREET_NAME, CALL_CENTER, CC_STREET_NAME},
    {0, 0, 0, 0, CC_STREET_TYPE, CALL_CENTER, CC_STREET_TYPE},
    {0, 0, 0, 0, CC_SUITE_NUMBER, CALL_CENTER, CC_SUITE_NUMBER},
    {0, 0, 0, 0, CC_CITY, CALL_CENTER, CC_CITY},
    {0, 0, 0, 0, CC_COUNTY, CALL_CENTER, CC_COUNTY},
    {0, 0, 0, 0, CC_STATE, CALL_CENTER, CC_STATE},
    {0, 0, 0, 0, CC_ZIP, CALL_CENTER, CC_ZIP},
    {0, 0, 0, 0, CC_COUNTRY, CALL_CENTER, CC_COUNTRY},
    {0, 0, 0, 0, CC_GMT_OFFSET, CALL_CENTER, CC_GMT_OFFSET},
    {0, 15, 0, 0, CC_ADDRESS, CALL_CENTER, CC_ADDRESS},
    {0, 1, 0, 0, CC_TAX_PERCENTAGE, CALL_CENTER, CC_TAX_PERCENTAGE},
    {0, 1, 0, 0, CC_SCD, CALL_CENTER, CC_SCD},
    {0, 2, 0, 0, CC_NULLS, CALL_CENTER, CC_NULLS},
    {0, 1, 0, 0, CP_CATALOG_PAGE_SK, CATALOG_PAGE, CP_CATALOG_PAGE_SK},
    {0, 1, 0, 0, CP_CATALOG_PAGE_ID, CATALOG_PAGE, CP_CATALOG_PAGE_ID},
    {0, 1, 0, 0, CP_START_DATE_ID, CATALOG_PAGE, CP_START_DATE_ID},
    {0, 1, 0, 0, CP_END_DATE_ID, CATALOG_PAGE, CP_END_DATE_ID},
    {0, 1, 0, 0, CP_PROMO_ID, CATALOG_PAGE, CP_PROMO_ID},
    {0, 1, 0, 0, CP_DEPARTMENT, CATALOG_PAGE, CP_DEPARTMENT},
    {0, 1, 0, 0, CP_CATALOG_NUMBER, CATALOG_PAGE, CP_CATALOG_NUMBER},
    {0, 1, 0, 0, CP_CATALOG_PAGE_NUMBER, CATALOG_PAGE, CP_CATALOG_PAGE_NUMBER},
    {0, 100, 0, 0, CP_DESCRIPTION, CATALOG_PAGE, S_CP_DESCRIPTION},
    {0, 1, 0, 0, CP_TYPE, CATALOG_PAGE, CP_TYPE},
    {0, 2, 0, 0, CP_NULLS, CATALOG_PAGE, CP_NULLS},
    {0, 28, 0, 0, CR_RETURNED_DATE_SK, CATALOG_RETURNS, CR_RETURNED_DATE_SK},
    {0, 28, 0, 0, CR_RETURNED_TIME_SK, CATALOG_RETURNS, CR_RETURNED_TIME_SK},
    {0, 14, 0, 0, CR_ITEM_SK, CATALOG_RETURNS, CR_ITEM_SK},
    {0, 14, 0, 0, CR_REFUNDED_CUSTOMER_SK, CATALOG_RETURNS, CR_REFUNDED_CUSTOMER_SK},
    {0, 14, 0, 0, CR_REFUNDED_CDEMO_SK, CATALOG_RETURNS, CR_REFUNDED_CDEMO_SK},
    {0, 14, 0, 0, CR_REFUNDED_HDEMO_SK, CATALOG_RETURNS, CR_REFUNDED_HDEMO_SK},
    {0, 14, 0, 0, CR_REFUNDED_ADDR_SK, CATALOG_RETURNS, CR_REFUNDED_ADDR_SK},
    {0, 28, 0, 0, CR_RETURNING_CUSTOMER_SK, CATALOG_RETURNS, CR_RETURNING_CUSTOMER_SK},
    {0, 14, 0, 0, CR_RETURNING_CDEMO_SK, CATALOG_RETURNS, CR_RETURNING_CDEMO_SK},
    {0, 14, 0, 0, CR_RETURNING_HDEMO_SK, CATALOG_RETURNS, CR_RETURNING_HDEMO_SK},
    {0, 14, 0, 0, CR_RETURNING_ADDR_SK, CATALOG_RETURNS, CR_RETURNING_ADDR_SK},
    {0, 0, 0, 0, CR_CALL_CENTER_SK, CATALOG_RETURNS, CR_CALL_CENTER_SK},
    {0, 14, 0, 0, CR_CATALOG_PAGE_SK, CATALOG_RETURNS, CR_CATALOG_PAGE_SK},
    {0, 14, 0, 0, CR_SHIP_MODE_SK, CATALOG_RETURNS, CR_SHIP_MODE_SK},
    {0, 14, 0, 0, CR_WAREHOUSE_SK, CATALOG_RETURNS, CR_WAREHOUSE_SK},
    {0, 14, 0, 0, CR_REASON_SK, CATALOG_RETURNS, CR_REASON_SK},
    {0, 0, 0, 0, CR_ORDER_NUMBER, CATALOG_RETURNS, CR_ORDER_NUMBER},
    {0, 0, 0, 0, CR_PRICING_QUANTITY, CATALOG_RETURNS, CR_PRICING_QUANTITY},
    {0, 0, 0, 0, CR_PRICING_NET_PAID, CATALOG_RETURNS, CR_PRICING_NET_PAID},
    {0, 0, 0, 0, CR_PRICING_EXT_TAX, CATALOG_RETURNS, CR_PRICING_EXT_TAX},
    {0, 0, 0, 0, CR_PRICING_NET_PAID_INC_TAX, CATALOG_RETURNS, CR_PRICING_NET_PAID_INC_TAX},
    {0, 0, 0, 0, CR_PRICING_FEE, CATALOG_RETURNS, CR_PRICING_FEE},
    {0, 0, 0, 0, CR_PRICING_EXT_SHIP_COST, CATALOG_RETURNS, CR_PRICING_EXT_SHIP_COST},
    {0, 0, 0, 0, CR_PRICING_REFUNDED_CASH, CATALOG_RETURNS, CR_PRICING_REFUNDED_CASH},
    {0, 0, 0, 0, CR_PRICING_REVERSED_CHARGE, CATALOG_RETURNS, CR_PRICING_REVERSED_CHARGE},
    {0, 0, 0, 0, CR_PRICING_STORE_CREDIT, CATALOG_RETURNS, CR_PRICING_STORE_CREDIT},
    {0, 0, 0, 0, CR_PRICING_NET_LOSS, CATALOG_RETURNS, CR_PRICING_NET_LOSS},
    {0, 28, 0, 0, CR_NULLS, CATALOG_RETURNS, CR_NULLS},
    {0, 70, 0, 0, CR_PRICING, CATALOG_RETURNS, CR_PRICING},
    {0, 1, 0, 0, CS_SOLD_DATE_SK, CATALOG_SALES, CS_SOLD_DATE_SK},
    {0, 2, 0, 0, CS_SOLD_TIME_SK, CATALOG_SALES, CS_SOLD_TIME_SK},
    {0, 14, 0, 0, CS_SHIP_DATE_SK, CATALOG_SALES, CS_SHIP_DATE_SK},
    {0, 1, 0, 0, CS_BILL_CUSTOMER_SK, CATALOG_SALES, CS_BILL_CUSTOMER_SK},
    {0, 1, 0, 0, CS_BILL_CDEMO_SK, CATALOG_SALES, CS_BILL_CDEMO_SK},
    {0, 1, 0, 0, CS_BILL_HDEMO_SK, CATALOG_SALES, CS_BILL_HDEMO_SK},
    {0, 1, 0, 0, CS_BILL_ADDR_SK, CATALOG_SALES, CS_BILL_ADDR_SK},
    {0, 2, 0, 0, CS_SHIP_CUSTOMER_SK, CATALOG_SALES, CS_SHIP_CUSTOMER_SK},
    {0, 1, 0, 0, CS_SHIP_CDEMO_SK, CATALOG_SALES, CS_SHIP_CDEMO_SK},
    {0, 1, 0, 0, CS_SHIP_HDEMO_SK, CATALOG_SALES, CS_SHIP_HDEMO_SK},
    {0, 1, 0, 0, CS_SHIP_ADDR_SK, CATALOG_SALES, CS_SHIP_ADDR_SK},
    {0, 1, 0, 0, CS_CALL_CENTER_SK, CATALOG_SALES, CS_CALL_CENTER_SK},
    {0, 42, 0, 0, CS_CATALOG_PAGE_SK, CATALOG_SALES, CS_CATALOG_PAGE_SK},
    {0, 14, 0, 0, CS_SHIP_MODE_SK, CATALOG_SALES, CS_SHIP_MODE_SK},
    {0, 14, 0, 0, CS_WAREHOUSE_SK, CATALOG_SALES, CS_WAREHOUSE_SK},
    {0, 1, 0, 0, CS_SOLD_ITEM_SK, CATALOG_SALES, CS_SOLD_ITEM_SK},
    {0, 14, 0, 0, CS_PROMO_SK, CATALOG_SALES, CS_PROMO_SK},
    {0, 1, 0, 0, CS_ORDER_NUMBER, CATALOG_SALES, CS_ORDER_NUMBER},
    {0, 0, 0, 0, CS_PRICING_QUANTITY, CATALOG_SALES, CS_PRICING_QUANTITY},
    {0, 0, 0, 0, CS_PRICING_WHOLESALE_COST, CATALOG_SALES, CS_PRICING_WHOLESALE_COST},
    {0, 0, 0, 0, CS_PRICING_LIST_PRICE, CATALOG_SALES, CS_PRICING_LIST_PRICE},
    {0, 0, 0, 0, CS_PRICING_SALES_PRICE, CATALOG_SALES, CS_PRICING_SALES_PRICE},
    {0, 0, 0, 0, CS_PRICING_COUPON_AMT, CATALOG_SALES, CS_PRICING_COUPON_AMT},
    {0, 0, 0, 0, CS_PRICING_EXT_SALES_PRICE, CATALOG_SALES, CS_PRICING_EXT_SALES_PRICE},
    {0, 0, 0, 0, CS_PRICING_EXT_DISCOUNT_AMOUNT, CATALOG_SALES, CS_PRICING_EXT_DISCOUNT_AMOUNT},
    {0, 0, 0, 0, CS_PRICING_EXT_WHOLESALE_COST, CATALOG_SALES, CS_PRICING_EXT_WHOLESALE_COST},
    {0, 0, 0, 0, CS_PRICING_EXT_LIST_PRICE, CATALOG_SALES, CS_PRICING_EXT_LIST_PRICE},
    {0, 0, 0, 0, CS_PRICING_EXT_TAX, CATALOG_SALES, CS_PRICING_EXT_TAX},
    {0, 0, 0, 0, CS_PRICING_EXT_SHIP_COST, CATALOG_SALES, CS_PRICING_EXT_SHIP_COST},
    {0, 0, 0, 0, CS_PRICING_NET_PAID, CATALOG_SALES, CS_PRICING_NET_PAID},
    {0, 0, 0, 0, CS_PRICING_NET_PAID_INC_TAX, CATALOG_SALES, CS_PRICING_NET_PAID_INC_TAX},
    {0, 0, 0, 0, CS_PRICING_NET_PAID_INC_SHIP, CATALOG_SALES, CS_PRICING_NET_PAID_INC_SHIP},
    {0, 0, 0, 0, CS_PRICING_NET_PAID_INC_SHIP_TAX, CATALOG_SALES, CS_PRICING_NET_PAID_INC_SHIP_TAX},
    {0, 0, 0, 0, CS_PRICING_NET_PROFIT, CATALOG_SALES, CS_PRICING_NET_PROFIT},
    {0, 112, 0, 0, CS_PRICING, CATALOG_SALES, CS_PRICING},
    {0, 0, 0, 0, CS_PERMUTE, CATALOG_SALES, CS_PERMUTE},
    {0, 28, 0, 0, CS_NULLS, CATALOG_SALES, CS_NULLS},
    {0, 14, 0, 0, CR_IS_RETURNED, CATALOG_SALES, CR_IS_RETURNED},
    {0, 0, 0, 0, CS_PERMUTATION, CATALOG_SALES, CS_PERMUTATION},
    {0, 1, 0, 0, C_CUSTOMER_SK, CUSTOMER, C_CUSTOMER_SK},
    {0, 1, 0, 0, C_CUSTOMER_ID, CUSTOMER, C_CUSTOMER_ID},
    {0, 1, 0, 0, C_CURRENT_CDEMO_SK, CUSTOMER, C_CURRENT_CDEMO_SK},
    {0, 1, 0, 0, C_CURRENT_HDEMO_SK, CUSTOMER, C_CURRENT_HDEMO_SK},
    {0, 1, 0, 0, C_CURRENT_ADDR_SK, CUSTOMER, C_CURRENT_ADDR_SK},
    {0, 0, 0, 0, C_FIRST_SHIPTO_DATE_ID, CUSTOMER, C_FIRST_SHIPTO_DATE_ID},
    {0, 1, 0, 0, C_FIRST_SALES_DATE_ID, CUSTOMER, C_FIRST_SALES_DATE_ID},
    {0, 1, 0, 0, C_SALUTATION, CUSTOMER, C_SALUTATION},
    {0, 1, 0, 0, C_FIRST_NAME, CUSTOMER, C_FIRST_NAME},
    {0, 1, 0, 0, C_LAST_NAME, CUSTOMER, C_LAST_NAME},
    {0, 2, 0, 0, C_PREFERRED_CUST_FLAG, CUSTOMER, C_PREFERRED_CUST_FLAG},
    {0, 1, 0, 0, C_BIRTH_DAY, CUSTOMER, C_BIRTH_DAY},
    {0, 0, 0, 0, C_BIRTH_MONTH, CUSTOMER, C_BIRTH_MONTH},
    {0, 0, 0, 0, C_BIRTH_YEAR, CUSTOMER, C_BIRTH_YEAR},
    {0, 1, 0, 0, C_BIRTH_COUNTRY, CUSTOMER, C_BIRTH_COUNTRY},
    {0, 1, 0, 0, C_LOGIN, CUSTOMER, C_LOGIN},
    {0, 23, 0, 0, C_EMAIL_ADDRESS, CUSTOMER, C_EMAIL_ADDRESS},
    {0, 1, 0, 0, C_LAST_REVIEW_DATE, CUSTOMER, C_LAST_REVIEW_DATE},
    {0, 2, 0, 0, C_NULLS, CUSTOMER, C_NULLS},
    {0, 1, 0, 0, CA_ADDRESS_SK, CUSTOMER_ADDRESS, CA_ADDRESS_SK},
    {0, 1, 0, 0, CA_ADDRESS_ID, CUSTOMER_ADDRESS, CA_ADDRESS_ID},
    {0, 1, 0, 0, CA_ADDRESS_STREET_NUM, CUSTOMER_ADDRESS, CA_ADDRESS_STREET_NUM},
    {0, 1, 0, 0, CA_ADDRESS_STREET_NAME1, CUSTOMER_ADDRESS, CA_ADDRESS_STREET_NAME1},
    {0, 1, 0, 0, CA_ADDRESS_STREET_TYPE, CUSTOMER_ADDRESS, CA_ADDRESS_STREET_TYPE},
    {0, 1, 0, 0, CA_ADDRESS_SUITE_NUM, CUSTOMER_ADDRESS, CA_ADDRESS_SUITE_NUM},
    {0, 1, 0, 0, CA_ADDRESS_CITY, CUSTOMER_ADDRESS, CA_ADDRESS_CITY},
    {0, 1, 0, 0, CA_ADDRESS_COUNTY, CUSTOMER_ADDRESS, CA_ADDRESS_COUNTY},
    {0, 1, 0, 0, CA_ADDRESS_STATE, CUSTOMER_ADDRESS, CA_ADDRESS_STATE},
    {0, 1, 0, 0, CA_ADDRESS_ZIP, CUSTOMER_ADDRESS, CA_ADDRESS_ZIP},
    {0, 1, 0, 0, CA_ADDRESS_COUNTRY, CUSTOMER_ADDRESS, CA_ADDRESS_COUNTRY},
    {0, 1, 0, 0, CA_ADDRESS_GMT_OFFSET, CUSTOMER_ADDRESS, CA_ADDRESS_GMT_OFFSET},
    {0, 1, 0, 0, CA_LOCATION_TYPE, CUSTOMER_ADDRESS, CA_LOCATION_TYPE},
    {0, 2, 0, 0, CA_NULLS, CUSTOMER_ADDRESS, CA_NULLS},
    {0, 7, 0, 0, CA_ADDRESS, CUSTOMER_ADDRESS, CA_ADDRESS},
    {0, 1, 0, 0, CA_ADDRESS_STREET_NAME2, CUSTOMER_ADDRESS, CA_ADDRESS_STREET_NAME2},
    {0, 1, 0, 0, CD_DEMO_SK, CUSTOMER_DEMOGRAPHICS, CD_DEMO_SK},
    {0, 1, 0, 0, CD_GENDER, CUSTOMER_DEMOGRAPHICS, CD_GENDER},
    {0, 1, 0, 0, CD_MARITAL_STATUS, CUSTOMER_DEMOGRAPHICS, CD_MARITAL_STATUS},
    {0, 1, 0, 0, CD_EDUCATION_STATUS, CUSTOMER_DEMOGRAPHICS, CD_EDUCATION_STATUS},
    {0, 1, 0, 0, CD_PURCHASE_ESTIMATE, CUSTOMER_DEMOGRAPHICS, CD_PURCHASE_ESTIMATE},
    {0, 1, 0, 0, CD_CREDIT_RATING, CUSTOMER_DEMOGRAPHICS, CD_CREDIT_RATING},
    {0, 1, 0, 0, CD_DEP_COUNT, CUSTOMER_DEMOGRAPHICS, CD_DEP_COUNT},
    {0, 1, 0, 0, CD_DEP_EMPLOYED_COUNT, CUSTOMER_DEMOGRAPHICS, CD_DEP_EMPLOYED_COUNT},
    {0, 1, 0, 0, CD_DEP_COLLEGE_COUNT, CUSTOMER_DEMOGRAPHICS, CD_DEP_COLLEGE_COUNT},
    {0, 2, 0, 0, CD_NULLS, CUSTOMER_DEMOGRAPHICS, CD_NULLS},
    {0, 0, 0, 0, D_DATE_SK, DATET, D_DATE_SK},
    {0, 0, 0, 0, D_DATE_ID, DATET, D_DATE_ID},
    {0, 0, 0, 0, D_DATE, DATET, D_DATE},
    {0, 0, 0, 0, D_MONTH_SEQ, DATET, D_MONTH_SEQ},
    {0, 0, 0, 0, D_WEEK_SEQ, DATET, D_WEEK_SEQ},
    {0, 0, 0, 0, D_QUARTER_SEQ, DATET, D_QUARTER_SEQ},
    {0, 0, 0, 0, D_YEAR, DATET, D_YEAR},
    {0, 0, 0, 0, D_DOW, DATET, D_DOW},
    {0, 0, 0, 0, D_MOY, DATET, D_MOY},
    {0, 0, 0, 0, D_DOM, DATET, D_DOM},
    {0, 0, 0, 0, D_QOY, DATET, D_QOY},
    {0, 0, 0, 0, D_FY_YEAR, DATET, D_FY_YEAR},
    {0, 0, 0, 0, D_FY_QUARTER_SEQ, DATET, D_FY_QUARTER_SEQ},
    {0, 0, 0, 0, D_FY_WEEK_SEQ, DATET, D_FY_WEEK_SEQ},
    {0, 0, 0, 0, D_DAY_NAME, DATET, D_DAY_NAME},
    {0, 0, 0, 0, D_QUARTER_NAME, DATET, D_QUARTER_NAME},
    {0, 0, 0, 0, D_HOLIDAY, DATET, D_HOLIDAY},
    {0, 0, 0, 0, D_WEEKEND, DATET, D_WEEKEND},
    {0, 0, 0, 0, D_FOLLOWING_HOLIDAY, DATET, D_FOLLOWING_HOLIDAY},
    {0, 0, 0, 0, D_FIRST_DOM, DATET, D_FIRST_DOM},
    {0, 0, 0, 0, D_LAST_DOM, DATET, D_LAST_DOM},
    {0, 0, 0, 0, D_SAME_DAY_LY, DATET, D_SAME_DAY_LY},
    {0, 0, 0, 0, D_SAME_DAY_LQ, DATET, D_SAME_DAY_LQ},
    {0, 0, 0, 0, D_CURRENT_DAY, DATET, D_CURRENT_DAY},
    {0, 0, 0, 0, D_CURRENT_WEEK, DATET, D_CURRENT_WEEK},
    {0, 0, 0, 0, D_CURRENT_MONTH, DATET, D_CURRENT_MONTH},
    {0, 0, 0, 0, D_CURRENT_QUARTER, DATET, D_CURRENT_QUARTER},
    {0, 0, 0, 0, D_CURRENT_YEAR, DATET, D_CURRENT_YEAR},
    {0, 2, 0, 0, D_NULLS, DATET, D_NULLS},
    {0, 1, 0, 0, HD_DEMO_SK, HOUSEHOLD_DEMOGRAPHICS, HD_DEMO_SK},
    {0, 1, 0, 0, HD_INCOME_BAND_ID, HOUSEHOLD_DEMOGRAPHICS, HD_INCOME_BAND_ID},
    {0, 1, 0, 0, HD_BUY_POTENTIAL, HOUSEHOLD_DEMOGRAPHICS, HD_BUY_POTENTIAL},
    {0, 1, 0, 0, HD_DEP_COUNT, HOUSEHOLD_DEMOGRAPHICS, HD_DEP_COUNT},
    {0, 1, 0, 0, HD_VEHICLE_COUNT, HOUSEHOLD_DEMOGRAPHICS, HD_VEHICLE_COUNT},
    {0, 2, 0, 0, HD_NULLS, HOUSEHOLD_DEMOGRAPHICS, HD_NULLS},
    {0, 1, 0, 0, IB_INCOME_BAND_ID, INCOME_BAND, IB_INCOME_BAND_ID},
    {0, 1, 0, 0, IB_LOWER_BOUND, INCOME_BAND, IB_LOWER_BOUND},
    {0, 1, 0, 0, IB_UPPER_BOUND, INCOME_BAND, IB_UPPER_BOUND},
    {0, 2, 0, 0, IB_NULLS, INCOME_BAND, IB_NULLS},
    {0, 1, 0, 0, INV_DATE_SK, INVENTORY, INV_DATE_SK},
    {0, 1, 0, 0, INV_ITEM_SK, INVENTORY, INV_ITEM_SK},
    {0, 1, 0, 0, INV_WAREHOUSE_SK, INVENTORY, INV_WAREHOUSE_SK},
    {0, 1, 0, 0, INV_QUANTITY_ON_HAND, INVENTORY, INV_QUANTITY_ON_HAND},
    {0, 2, 0, 0, INV_NULLS, INVENTORY, INV_NULLS},
    {0, 1, 0, 0, I_ITEM_SK, ITEM, I_ITEM_SK},
    {0, 1, 0, 0, I_ITEM_ID, ITEM, I_ITEM_ID},
    {0, 1, 0, 0, I_REC_START_DATE_ID, ITEM, I_REC_START_DATE_ID},
    {0, 2, 0, 0, I_REC_END_DATE_ID, ITEM, I_REC_END_DATE_ID},
    {0, 200, 0, 0, I_ITEM_DESC, ITEM, S_ITEM_DESC},
    {0, 2, 0, 0, I_CURRENT_PRICE, ITEM, I_CURRENT_PRICE},
    {0, 1, 0, 0, I_WHOLESALE_COST, ITEM, I_WHOLESALE_COST},
    {0, 1, 0, 0, I_BRAND_ID, ITEM, I_BRAND_ID},
    {0, 1, 0, 0, I_BRAND, ITEM, I_BRAND},
    {0, 1, 0, 0, I_CLASS_ID, ITEM, I_CLASS_ID},
    {0, 1, 0, 0, I_CLASS, ITEM, I_CLASS},
    {0, 1, 0, 0, I_CATEGORY_ID, ITEM, I_CATEGORY_ID},
    {0, 1, 0, 0, I_CATEGORY, ITEM, I_CATEGORY},
    {0, 2, 0, 0, I_MANUFACT_ID, ITEM, I_MANUFACT_ID},
    {0, 1, 0, 0, I_MANUFACT, ITEM, I_MANUFACT},
    {0, 1, 0, 0, I_SIZE, ITEM, I_SIZE},
    {0, 50, 0, 0, I_FORMULATION, ITEM, I_FORMULATION},
    {0, 1, 0, 0, I_COLOR, ITEM, I_COLOR},
    {0, 1, 0, 0, I_UNITS, ITEM, I_UNITS},
    {0, 1, 0, 0, I_CONTAINER, ITEM, I_CONTAINER},
    {0, 2, 0, 0, I_MANAGER_ID, ITEM, I_MANAGER_ID},
    {0, 1, 0, 0, I_PRODUCT_NAME, ITEM, I_PRODUCT_NAME},
    {0, 2, 0, 0, I_NULLS, ITEM, I_NULLS},
    {0, 1, 0, 0, I_SCD, ITEM, I_SCD},
    {0, 2, 0, 0, I_PROMO_SK, ITEM, I_PROMO_SK},
    {0, 1, 0, 0, P_PROMO_SK, PROMOTION, P_PROMO_SK},
    {0, 1, 0, 0, P_PROMO_ID, PROMOTION, P_PROMO_ID},
    {0, 1, 0, 0, P_START_DATE_ID, PROMOTION, P_START_DATE_ID},
    {0, 1, 0, 0, P_END_DATE_ID, PROMOTION, P_END_DATE_ID},
    {0, 1, 0, 0, P_ITEM_SK, PROMOTION, P_ITEM_SK},
    {0, 1, 0, 0, P_COST, PROMOTION, P_COST},
    {0, 1, 0, 0, P_RESPONSE_TARGET, PROMOTION, P_RESPONSE_TARGET},
    {0, 1, 0, 0, P_PROMO_NAME, PROMOTION, P_PROMO_NAME},
    {0, 1, 0, 0, P_CHANNEL_DMAIL, PROMOTION, P_CHANNEL_DMAIL},
    {0, 1, 0, 0, P_CHANNEL_EMAIL, PROMOTION, P_CHANNEL_EMAIL},
    {0, 1, 0, 0, P_CHANNEL_CATALOG, PROMOTION, P_CHANNEL_CATALOG},
    {0, 1, 0, 0, P_CHANNEL_TV, PROMOTION, P_CHANNEL_TV},
    {0, 1, 0, 0, P_CHANNEL_RADIO, PROMOTION, P_CHANNEL_RADIO},
    {0, 1, 0, 0, P_CHANNEL_PRESS, PROMOTION, P_CHANNEL_PRESS},
    {0, 1, 0, 0, P_CHANNEL_EVENT, PROMOTION, P_CHANNEL_EVENT},
    {0, 1, 0, 0, P_CHANNEL_DEMO, PROMOTION, P_CHANNEL_DEMO},
    {0, 100, 0, 0, P_CHANNEL_DETAILS, PROMOTION, P_CHANNEL_DETAILS},
    {0, 1, 0, 0, P_PURPOSE, PROMOTION, P_PURPOSE},
    {0, 1, 0, 0, P_DISCOUNT_ACTIVE, PROMOTION, P_DISCOUNT_ACTIVE},
    {0, 2, 0, 0, P_NULLS, PROMOTION, P_NULLS},
    {0, 1, 0, 0, R_REASON_SK, REASON, R_REASON_SK},
    {0, 1, 0, 0, R_REASON_ID, REASON, R_REASON_ID},
    {0, 1, 0, 0, R_REASON_DESCRIPTION, REASON, R_REASON_DESCRIPTION},
    {0, 2, 0, 0, R_NULLS, REASON, R_NULLS},
    {0, 1, 0, 0, SM_SHIP_MODE_SK, SHIP_MODE, SM_SHIP_MODE_SK},
    {0, 1, 0, 0, SM_SHIP_MODE_ID, SHIP_MODE, SM_SHIP_MODE_ID},
    {0, 1, 0, 0, SM_TYPE, SHIP_MODE, SM_TYPE},
    {0, 1, 0, 0, SM_CODE, SHIP_MODE, SM_CODE},
    {0, 21, 0, 0, SM_CONTRACT, SHIP_MODE, SM_CONTRACT},
    {0, 1, 0, 0, SM_CARRIER, SHIP_MODE, SM_CARRIER},
    {0, 2, 0, 0, SM_NULLS, SHIP_MODE, SM_NULLS},
    {0, 1, 0, 0, W_STORE_SK, STORE, W_STORE_SK},
    {0, 1, 0, 0, W_STORE_ID, STORE, W_STORE_ID},
    {0, 1, 0, 0, W_STORE_REC_START_DATE_ID, STORE, W_STORE_REC_START_DATE_ID},
    {0, 2, 0, 0, W_STORE_REC_END_DATE_ID, STORE, W_STORE_REC_END_DATE_ID},
    {0, 2, 0, 0, W_STORE_CLOSED_DATE_ID, STORE, W_STORE_CLOSED_DATE_ID},
    {0, 0, 0, 0, W_STORE_NAME, STORE, W_STORE_NAME},
    {0, 1, 0, 0, W_STORE_EMPLOYEES, STORE, W_STORE_EMPLOYEES},
    {0, 1, 0, 0, W_STORE_FLOOR_SPACE, STORE, W_STORE_FLOOR_SPACE},
    {0, 1, 0, 0, W_STORE_HOURS, STORE, W_STORE_HOURS},
    {0, 2, 0, 0, W_STORE_MANAGER, STORE, W_STORE_MANAGER},
    {0, 1, 0, 0, W_STORE_MARKET_ID, STORE, W_STORE_MARKET_ID},
    {0, 1, 0, 0, W_STORE_TAX_PERCENTAGE, STORE, W_STORE_TAX_PERCENTAGE},
    {0, 1, 0, 0, W_STORE_GEOGRAPHY_CLASS, STORE, W_STORE_GEOGRAPHY_CLASS},
    {0, 100, 0, 0, W_STORE_MARKET_DESC, STORE, W_STORE_MARKET_DESC},
    {0, 2, 0, 0, W_STORE_MARKET_MANAGER, STORE, W_STORE_MARKET_MANAGER},
    {0, 1, 0, 0, W_STORE_DIVISION_ID, STORE, W_STORE_DIVISION_ID},
    {0, 1, 0, 0, W_STORE_DIVISION_NAME, STORE, W_STORE_DIVISION_NAME},
    {0, 1, 0, 0, W_STORE_COMPANY_ID, STORE, W_STORE_COMPANY_ID},
    {0, 1, 0, 0, W_STORE_COMPANY_NAME, STORE, W_STORE_COMPANY_NAME},
    {0, 1, 0, 0, W_STORE_ADDRESS_STREET_NUM, STORE, W_STORE_ADDRESS_STREET_NUM},
    {0, 1, 0, 0, W_STORE_ADDRESS_STREET_NAME1, STORE, W_STORE_ADDRESS_STREET_NAME1},
    {0, 1, 0, 0, W_STORE_ADDRESS_STREET_TYPE, STORE, W_STORE_ADDRESS_STREET_TYPE},
    {0, 1, 0, 0, W_STORE_ADDRESS_SUITE_NUM, STORE, W_STORE_ADDRESS_SUITE_NUM},
    {0, 1, 0, 0, W_STORE_ADDRESS_CITY, STORE, W_STORE_ADDRESS_CITY},
    {0, 1, 0, 0, W_STORE_ADDRESS_COUNTY, STORE, W_STORE_ADDRESS_COUNTY},
    {0, 1, 0, 0, W_STORE_ADDRESS_STATE, STORE, W_STORE_ADDRESS_STATE},
    {0, 1, 0, 0, W_STORE_ADDRESS_ZIP, STORE, W_STORE_ADDRESS_ZIP},
    {0, 1, 0, 0, W_STORE_ADDRESS_COUNTRY, STORE, W_STORE_ADDRESS_COUNTRY},
    {0, 1, 0, 0, W_STORE_ADDRESS_GMT_OFFSET, STORE, W_STORE_ADDRESS_GMT_OFFSET},
    {0, 2, 0, 0, W_STORE_NULLS, STORE, W_STORE_NULLS},
    {0, 1, 0, 0, W_STORE_TYPE, STORE, W_STORE_TYPE},
    {0, 1, 0, 0, W_STORE_SCD, STORE, W_STORE_SCD},
    {0, 7, 0, 0, W_STORE_ADDRESS, STORE, W_STORE_ADDRESS},
    {0, 32, 0, 0, SR_RETURNED_DATE_SK, STORE_RETURNS, SR_RETURNED_DATE_SK},
    {0, 32, 0, 0, SR_RETURNED_TIME_SK, STORE_RETURNS, SR_RETURNED_TIME_SK},
    {0, 16, 0, 0, SR_ITEM_SK, STORE_RETURNS, SR_ITEM_SK},
    {0, 16, 0, 0, SR_CUSTOMER_SK, STORE_RETURNS, SR_CUSTOMER_SK},
    {0, 16, 0, 0, SR_CDEMO_SK, STORE_RETURNS, SR_CDEMO_SK},
    {0, 16, 0, 0, SR_HDEMO_SK, STORE_RETURNS, SR_HDEMO_SK},
    {0, 16, 0, 0, SR_ADDR_SK, STORE_RETURNS, SR_ADDR_SK},
    {0, 16, 0, 0, SR_STORE_SK, STORE_RETURNS, SR_STORE_SK},
    {0, 16, 0, 0, SR_REASON_SK, STORE_RETURNS, SR_REASON_SK},
    {0, 16, 0, 0, SR_TICKET_NUMBER, STORE_RETURNS, SR_TICKET_NUMBER},
    {0, 0, 0, 0, SR_PRICING_QUANTITY, STORE_RETURNS, SR_PRICING_QUANTITY},
    {0, 0, 0, 0, SR_PRICING_NET_PAID, STORE_RETURNS, SR_PRICING_NET_PAID},
    {0, 0, 0, 0, SR_PRICING_EXT_TAX, STORE_RETURNS, SR_PRICING_EXT_TAX},
    {0, 0, 0, 0, SR_PRICING_NET_PAID_INC_TAX, STORE_RETURNS, SR_PRICING_NET_PAID_INC_TAX},
    {0, 0, 0, 0, SR_PRICING_FEE, STORE_RETURNS, SR_PRICING_FEE},
    {0, 0, 0, 0, SR_PRICING_EXT_SHIP_COST, STORE_RETURNS, SR_PRICING_EXT_SHIP_COST},
    {0, 0, 0, 0, SR_PRICING_REFUNDED_CASH, STORE_RETURNS, SR_PRICING_REFUNDED_CASH},
    {0, 0, 0, 0, SR_PRICING_REVERSED_CHARGE, STORE_RETURNS, SR_PRICING_REVERSED_CHARGE},
    {0, 0, 0, 0, SR_PRICING_STORE_CREDIT, STORE_RETURNS, SR_PRICING_STORE_CREDIT},
    {0, 0, 0, 0, SR_PRICING_NET_LOSS, STORE_RETURNS, SR_PRICING_NET_LOSS},
    {0, 80, 0, 0, SR_PRICING, STORE_RETURNS, SR_PRICING},
    {0, 32, 0, 0, SR_NULLS, STORE_RETURNS, SR_NULLS},
    {0, 2, 0, 0, SS_SOLD_DATE_SK, STORE_SALES, SS_SOLD_DATE_SK},
    {0, 2, 0, 0, SS_SOLD_TIME_SK, STORE_SALES, SS_SOLD_TIME_SK},
    {0, 1, 0, 0, SS_SOLD_ITEM_SK, STORE_SALES, SS_SOLD_ITEM_SK},
    {0, 1, 0, 0, SS_SOLD_CUSTOMER_SK, STORE_SALES, SS_SOLD_CUSTOMER_SK},
    {0, 1, 0, 0, SS_SOLD_CDEMO_SK, STORE_SALES, SS_SOLD_CDEMO_SK},
    {0, 1, 0, 0, SS_SOLD_HDEMO_SK, STORE_SALES, SS_SOLD_HDEMO_SK},
    {0, 1, 0, 0, SS_SOLD_ADDR_SK, STORE_SALES, SS_SOLD_ADDR_SK},
    {0, 1, 0, 0, SS_SOLD_STORE_SK, STORE_SALES, SS_SOLD_STORE_SK},
    {0, 16, 0, 0, SS_SOLD_PROMO_SK, STORE_SALES, SS_SOLD_PROMO_SK},
    {0, 1, 0, 0, SS_TICKET_NUMBER, STORE_SALES, SS_TICKET_NUMBER},
    {0, 1, 0, 0, SS_PRICING_QUANTITY, STORE_SALES, SS_PRICING_QUANTITY},
    {0, 0, 0, 0, SS_PRICING_WHOLESALE_COST, STORE_SALES, SS_PRICING_WHOLESALE_COST},
    {0, 0, 0, 0, SS_PRICING_LIST_PRICE, STORE_SALES, SS_PRICING_LIST_PRICE},
    {0, 0, 0, 0, SS_PRICING_SALES_PRICE, STORE_SALES, SS_PRICING_SALES_PRICE},
    {0, 0, 0, 0, SS_PRICING_COUPON_AMT, STORE_SALES, SS_PRICING_COUPON_AMT},
    {0, 0, 0, 0, SS_PRICING_EXT_SALES_PRICE, STORE_SALES, SS_PRICING_EXT_SALES_PRICE},
    {0, 0, 0, 0, SS_PRICING_EXT_WHOLESALE_COST, STORE_SALES, SS_PRICING_EXT_WHOLESALE_COST},
    {0, 0, 0, 0, SS_PRICING_EXT_LIST_PRICE, STORE_SALES, SS_PRICING_EXT_LIST_PRICE},
    {0, 0, 0, 0, SS_PRICING_EXT_TAX, STORE_SALES, SS_PRICING_EXT_TAX},
    {0, 0, 0, 0, SS_PRICING_NET_PAID, STORE_SALES, SS_PRICING_NET_PAID},
    {0, 0, 0, 0, SS_PRICING_NET_PAID_INC_TAX, STORE_SALES, SS_PRICING_NET_PAID_INC_TAX},
    {0, 0, 0, 0, SS_PRICING_NET_PROFIT, STORE_SALES, SS_PRICING_NET_PROFIT},
    {0, 16, 0, 0, SR_IS_RETURNED, STORE_SALES, SR_IS_RETURNED},
    {0, 128, 0, 0, SS_PRICING, STORE_SALES, SS_PRICING},
    {0, 32, 0, 0, SS_NULLS, STORE_SALES, SS_NULLS},
    {0, 0, 0, 0, SS_PERMUTATION, STORE_SALES, SS_PERMUTATION},
    {0, 1, 0, 0, T_TIME_SK, TIME, T_TIME_SK},
    {0, 1, 0, 0, T_TIME_ID, TIME, T_TIME_ID},
    {0, 1, 0, 0, T_TIME, TIME, T_TIME},
    {0, 1, 0, 0, T_HOUR, TIME, T_HOUR},
    {0, 1, 0, 0, T_MINUTE, TIME, T_MINUTE},
    {0, 1, 0, 0, T_SECOND, TIME, T_SECOND},
    {0, 1, 0, 0, T_AM_PM, TIME, T_AM_PM},
    {0, 1, 0, 0, T_SHIFT, TIME, T_SHIFT},
    {0, 1, 0, 0, T_SUB_SHIFT, TIME, T_SUB_SHIFT},
    {0, 1, 0, 0, T_MEAL_TIME, TIME, T_MEAL_TIME},
    {0, 2, 0, 0, T_NULLS, TIME, T_NULLS},
    {0, 1, 0, 0, W_WAREHOUSE_SK, WAREHOUSE, W_WAREHOUSE_SK},
    {0, 1, 0, 0, W_WAREHOUSE_ID, WAREHOUSE, W_WAREHOUSE_ID},
    {0, 80, 0, 0, W_WAREHOUSE_NAME, WAREHOUSE, W_WAREHOUSE_NAME},
    {0, 1, 0, 0, W_WAREHOUSE_SQ_FT, WAREHOUSE, W_WAREHOUSE_SQ_FT},
    {0, 1, 0, 0, W_ADDRESS_STREET_NUM, WAREHOUSE, W_ADDRESS_STREET_NUM},
    {0, 1, 0, 0, W_ADDRESS_STREET_NAME1, WAREHOUSE, W_ADDRESS_STREET_NAME1},
    {0, 1, 0, 0, W_ADDRESS_STREET_TYPE, WAREHOUSE, W_ADDRESS_STREET_TYPE},
    {0, 1, 0, 0, W_ADDRESS_SUITE_NUM, WAREHOUSE, W_ADDRESS_SUITE_NUM},
    {0, 1, 0, 0, W_ADDRESS_CITY, WAREHOUSE, W_ADDRESS_CITY},
    {0, 1, 0, 0, W_ADDRESS_COUNTY, WAREHOUSE, W_ADDRESS_COUNTY},
    {0, 1, 0, 0, W_ADDRESS_STATE, WAREHOUSE, W_ADDRESS_STATE},
    {0, 1, 0, 0, W_ADDRESS_ZIP, WAREHOUSE, W_ADDRESS_ZIP},
    {0, 1, 0, 0, W_ADDRESS_COUNTRY, WAREHOUSE, W_ADDRESS_COUNTRY},
    {0, 1, 0, 0, W_ADDRESS_GMT_OFFSET, WAREHOUSE, W_ADDRESS_GMT_OFFSET},
    {0, 2, 0, 0, W_NULLS, WAREHOUSE, W_NULLS},
    {0, 7, 0, 0, W_WAREHOUSE_ADDRESS, WAREHOUSE, W_WAREHOUSE_ADDRESS},
    {0, 1, 0, 0, WP_PAGE_SK, WEB_PAGE, WP_PAGE_SK},
    {0, 1, 0, 0, WP_PAGE_ID, WEB_PAGE, WP_PAGE_ID},
    {0, 1, 0, 0, WP_REC_START_DATE_ID, WEB_PAGE, WP_REC_START_DATE_ID},
    {0, 1, 0, 0, WP_REC_END_DATE_ID, WEB_PAGE, WP_REC_END_DATE_ID},
    {0, 2, 0, 0, WP_CREATION_DATE_SK, WEB_PAGE, WP_CREATION_DATE_SK},
    {0, 1, 0, 0, WP_ACCESS_DATE_SK, WEB_PAGE, WP_ACCESS_DATE_SK},
    {0, 1, 0, 0, WP_AUTOGEN_FLAG, WEB_PAGE, WP_AUTOGEN_FLAG},
    {0, 1, 0, 0, WP_CUSTOMER_SK, WEB_PAGE, WP_CUSTOMER_SK},
    {0, 1, 0, 0, WP_URL, WEB_PAGE, WP_URL},
    {0, 1, 0, 0, WP_TYPE, WEB_PAGE, WP_TYPE},
    {0, 1, 0, 0, WP_CHAR_COUNT, WEB_PAGE, WP_CHAR_COUNT},
    {0, 1, 0, 0, WP_LINK_COUNT, WEB_PAGE, WP_LINK_COUNT},
    {0, 1, 0, 0, WP_IMAGE_COUNT, WEB_PAGE, WP_IMAGE_COUNT},
    {0, 1, 0, 0, WP_MAX_AD_COUNT, WEB_PAGE, WP_MAX_AD_COUNT},
    {0, 2, 0, 0, WP_NULLS, WEB_PAGE, WP_NULLS},
    {0, 1, 0, 0, WP_SCD, WEB_PAGE, WP_SCD},
    {0, 32, 0, 0, WR_RETURNED_DATE_SK, WEB_RETURNS, WR_RETURNED_DATE_SK},
    {0, 32, 0, 0, WR_RETURNED_TIME_SK, WEB_RETURNS, WR_RETURNED_TIME_SK},
    {0, 16, 0, 0, WR_ITEM_SK, WEB_RETURNS, WR_ITEM_SK},
    {0, 16, 0, 0, WR_REFUNDED_CUSTOMER_SK, WEB_RETURNS, WR_REFUNDED_CUSTOMER_SK},
    {0, 16, 0, 0, WR_REFUNDED_CDEMO_SK, WEB_RETURNS, WR_REFUNDED_CDEMO_SK},
    {0, 16, 0, 0, WR_REFUNDED_HDEMO_SK, WEB_RETURNS, WR_REFUNDED_HDEMO_SK},
    {0, 16, 0, 0, WR_REFUNDED_ADDR_SK, WEB_RETURNS, WR_REFUNDED_ADDR_SK},
    {0, 16, 0, 0, WR_RETURNING_CUSTOMER_SK, WEB_RETURNS, WR_RETURNING_CUSTOMER_SK},
    {0, 16, 0, 0, WR_RETURNING_CDEMO_SK, WEB_RETURNS, WR_RETURNING_CDEMO_SK},
    {0, 16, 0, 0, WR_RETURNING_HDEMO_SK, WEB_RETURNS, WR_RETURNING_HDEMO_SK},
    {0, 16, 0, 0, WR_RETURNING_ADDR_SK, WEB_RETURNS, WR_RETURNING_ADDR_SK},
    {0, 16, 0, 0, WR_WEB_PAGE_SK, WEB_RETURNS, WR_WEB_PAGE_SK},
    {0, 16, 0, 0, WR_REASON_SK, WEB_RETURNS, WR_REASON_SK},
    {0, 0, 0, 0, WR_ORDER_NUMBER, WEB_RETURNS, WR_ORDER_NUMBER},
    {0, 0, 0, 0, WR_PRICING_QUANTITY, WEB_RETURNS, WR_PRICING_QUANTITY},
    {0, 0, 0, 0, WR_PRICING_NET_PAID, WEB_RETURNS, WR_PRICING_NET_PAID},
    {0, 0, 0, 0, WR_PRICING_EXT_TAX, WEB_RETURNS, WR_PRICING_EXT_TAX},
    {0, 0, 0, 0, WR_PRICING_NET_PAID_INC_TAX, WEB_RETURNS, WR_PRICING_NET_PAID_INC_TAX},
    {0, 0, 0, 0, WR_PRICING_FEE, WEB_RETURNS, WR_PRICING_FEE},
    {0, 0, 0, 0, WR_PRICING_EXT_SHIP_COST, WEB_RETURNS, WR_PRICING_EXT_SHIP_COST},
    {0, 0, 0, 0, WR_PRICING_REFUNDED_CASH, WEB_RETURNS, WR_PRICING_REFUNDED_CASH},
    {0, 0, 0, 0, WR_PRICING_REVERSED_CHARGE, WEB_RETURNS, WR_PRICING_REVERSED_CHARGE},
    {0, 0, 0, 0, WR_PRICING_STORE_CREDIT, WEB_RETURNS, WR_PRICING_STORE_CREDIT},
    {0, 0, 0, 0, WR_PRICING_NET_LOSS, WEB_RETURNS, WR_PRICING_NET_LOSS},
    {0, 80, 0, 0, WR_PRICING, WEB_RETURNS, WR_PRICING},
    {0, 32, 0, 0, WR_NULLS, WEB_RETURNS, WR_NULLS},
    {0, 2, 0, 0, WS_SOLD_DATE_SK, WEB_SALES, WS_SOLD_DATE_SK},
    {0, 2, 0, 0, WS_SOLD_TIME_SK, WEB_SALES, WS_SOLD_TIME_SK},
    {0, 16, 0, 0, WS_SHIP_DATE_SK, WEB_SALES, WS_SHIP_DATE_SK},
    {0, 1, 0, 0, WS_ITEM_SK, WEB_SALES, WS_ITEM_SK},
    {0, 1, 0, 0, WS_BILL_CUSTOMER_SK, WEB_SALES, WS_BILL_CUSTOMER_SK},
    {0, 1, 0, 0, WS_BILL_CDEMO_SK, WEB_SALES, WS_BILL_CDEMO_SK},
    {0, 1, 0, 0, WS_BILL_HDEMO_SK, WEB_SALES, WS_BILL_HDEMO_SK},
    {0, 1, 0, 0, WS_BILL_ADDR_SK, WEB_SALES, WS_BILL_ADDR_SK},
    {0, 2, 0, 0, WS_SHIP_CUSTOMER_SK, WEB_SALES, WS_SHIP_CUSTOMER_SK},
    {0, 2, 0, 0, WS_SHIP_CDEMO_SK, WEB_SALES, WS_SHIP_CDEMO_SK},
    {0, 1, 0, 0, WS_SHIP_HDEMO_SK, WEB_SALES, WS_SHIP_HDEMO_SK},
    {0, 1, 0, 0, WS_SHIP_ADDR_SK, WEB_SALES, WS_SHIP_ADDR_SK},
    {0, 16, 0, 0, WS_WEB_PAGE_SK, WEB_SALES, WS_WEB_PAGE_SK},
    {0, 16, 0, 0, WS_WEB_SITE_SK, WEB_SALES, WS_WEB_SITE_SK},
    {0, 16, 0, 0, WS_SHIP_MODE_SK, WEB_SALES, WS_SHIP_MODE_SK},
    {0, 16, 0, 0, WS_WAREHOUSE_SK, WEB_SALES, WS_WAREHOUSE_SK},
    {0, 16, 0, 0, WS_PROMO_SK, WEB_SALES, WS_PROMO_SK},
    {0, 1, 0, 0, WS_ORDER_NUMBER, WEB_SALES, WS_ORDER_NUMBER},
    {0, 1, 0, 0, WS_PRICING_QUANTITY, WEB_SALES, WS_PRICING_QUANTITY},
    {0, 1, 0, 0, WS_PRICING_WHOLESALE_COST, WEB_SALES, WS_PRICING_WHOLESALE_COST},
    {0, 0, 0, 0, WS_PRICING_LIST_PRICE, WEB_SALES, WS_PRICING_LIST_PRICE},
    {0, 0, 0, 0, WS_PRICING_SALES_PRICE, WEB_SALES, WS_PRICING_SALES_PRICE},
    {0, 0, 0, 0, WS_PRICING_EXT_DISCOUNT_AMT, WEB_SALES, WS_PRICING_EXT_DISCOUNT_AMT},
    {0, 0, 0, 0, WS_PRICING_EXT_SALES_PRICE, WEB_SALES, WS_PRICING_EXT_SALES_PRICE},
    {0, 0, 0, 0, WS_PRICING_EXT_WHOLESALE_COST, WEB_SALES, WS_PRICING_EXT_WHOLESALE_COST},
    {0, 0, 0, 0, WS_PRICING_EXT_LIST_PRICE, WEB_SALES, WS_PRICING_EXT_LIST_PRICE},
    {0, 0, 0, 0, WS_PRICING_EXT_TAX, WEB_SALES, WS_PRICING_EXT_TAX},
    {0, 0, 0, 0, WS_PRICING_COUPON_AMT, WEB_SALES, WS_PRICING_COUPON_AMT},
    {0, 0, 0, 0, WS_PRICING_EXT_SHIP_COST, WEB_SALES, WS_PRICING_EXT_SHIP_COST},
    {0, 0, 0, 0, WS_PRICING_NET_PAID, WEB_SALES, WS_PRICING_NET_PAID},
    {0, 0, 0, 0, WS_PRICING_NET_PAID_INC_TAX, WEB_SALES, WS_PRICING_NET_PAID_INC_TAX},
    {0, 0, 0, 0, WS_PRICING_NET_PAID_INC_SHIP, WEB_SALES, WS_PRICING_NET_PAID_INC_SHIP},
    {0, 0, 0, 0, WS_PRICING_NET_PAID_INC_SHIP_TAX, WEB_SALES, WS_PRICING_NET_PAID_INC_SHIP_TAX},
    {0, 0, 0, 0, WS_PRICING_NET_PROFIT, WEB_SALES, WS_PRICING_NET_PROFIT},
    {0, 128, 0, 0, WS_PRICING, WEB_SALES, WS_PRICING},
    {0, 32, 0, 0, WS_NULLS, WEB_SALES, WS_NULLS},
    {0, 16, 0, 0, WR_IS_RETURNED, WEB_SALES, WR_IS_RETURNED},
    {0, 0, 0, 0, WS_PERMUTATION, WEB_SALES, WS_PERMUTATION},
    {0, 1, 0, 0, WEB_SITE_SK, WEB_SITE, WEB_SITE_SK},
    {0, 1, 0, 0, WEB_SITE_ID, WEB_SITE, WEB_SITE_ID},
    {0, 1, 0, 0, WEB_REC_START_DATE_ID, WEB_SITE, WEB_REC_START_DATE_ID},
    {0, 1, 0, 0, WEB_REC_END_DATE_ID, WEB_SITE, WEB_REC_END_DATE_ID},
    {0, 1, 0, 0, WEB_NAME, WEB_SITE, WEB_NAME},
    {0, 1, 0, 0, WEB_OPEN_DATE, WEB_SITE, WEB_OPEN_DATE},
    {0, 1, 0, 0, WEB_CLOSE_DATE, WEB_SITE, WEB_CLOSE_DATE},
    {0, 1, 0, 0, WEB_CLASS, WEB_SITE, WEB_CLASS},
    {0, 2, 0, 0, WEB_MANAGER, WEB_SITE, WEB_MANAGER},
    {0, 1, 0, 0, WEB_MARKET_ID, WEB_SITE, WEB_MARKET_ID},
    {0, 20, 0, 0, WEB_MARKET_CLASS, WEB_SITE, WEB_MARKET_CLASS},
    {0, 100, 0, 0, WEB_MARKET_DESC, WEB_SITE, WEB_MARKET_DESC},
    {0, 2, 0, 0, WEB_MARKET_MANAGER, WEB_SITE, WEB_MARKET_MANAGER},
    {0, 1, 0, 0, WEB_COMPANY_ID, WEB_SITE, WEB_COMPANY_ID},
    {0, 1, 0, 0, WEB_COMPANY_NAME, WEB_SITE, WEB_COMPANY_NAME},
    {0, 1, 0, 0, WEB_ADDRESS_STREET_NUM, WEB_SITE, WEB_ADDRESS_STREET_NUM},
    {0, 1, 0, 0, WEB_ADDRESS_STREET_NAME1, WEB_SITE, WEB_ADDRESS_STREET_NAME1},
    {0, 1, 0, 0, WEB_ADDRESS_STREET_TYPE, WEB_SITE, WEB_ADDRESS_STREET_TYPE},
    {0, 1, 0, 0, WEB_ADDRESS_SUITE_NUM, WEB_SITE, WEB_ADDRESS_SUITE_NUM},
    {0, 1, 0, 0, WEB_ADDRESS_CITY, WEB_SITE, WEB_ADDRESS_CITY},
    {0, 1, 0, 0, WEB_ADDRESS_COUNTY, WEB_SITE, WEB_ADDRESS_COUNTY},
    {0, 1, 0, 0, WEB_ADDRESS_STATE, WEB_SITE, WEB_ADDRESS_STATE},
    {0, 1, 0, 0, WEB_ADDRESS_ZIP, WEB_SITE, WEB_ADDRESS_ZIP},
    {0, 1, 0, 0, WEB_ADDRESS_COUNTRY, WEB_SITE, WEB_ADDRESS_COUNTRY},
    {0, 1, 0, 0, WEB_ADDRESS_GMT_OFFSET, WEB_SITE, WEB_ADDRESS_GMT_OFFSET},
    {0, 1, 0, 0, WEB_TAX_PERCENTAGE, WEB_SITE, WEB_TAX_PERCENTAGE},
    {0, 2, 0, 0, WEB_NULLS, WEB_SITE, WEB_NULLS},
    {0, 7, 0, 0, WEB_ADDRESS, WEB_SITE, WEB_ADDRESS},
    {0, 70, 0, 0, WEB_SCD, WEB_SITE, WEB_SCD},
    {0, 0, 0, 0, DV_VERSION, DBGEN_VERSION, DV_VERSION},
    {0, 0, 0, 0, DV_CREATE_DATE, DBGEN_VERSION, DV_CREATE_DATE},
    {0, 0, 0, 0, DV_CREATE_TIME, DBGEN_VERSION, DV_CREATE_TIME},
    {0, 0, 0, 0, DV_CMDLINE_ARGS, DBGEN_VERSION, DV_CMDLINE_ARGS},
    {0, 0, 0, 0, VALIDATE_STREAM, DBGEN_VERSION, VALIDATE_STREAM},
    {0, 0, 0, 0, S_BRAND_ID, S_BRAND, S_BRAND_ID},
    {0, 0, 0, 0, S_BRAND_SUBCLASS_ID, S_BRAND, S_BRAND_SUBCLASS_ID},
    {0, 1, 0, 0, S_BRAND_MANAGER_ID, S_BRAND, S_BRAND_MANAGER_ID},
    {0, 1, 0, 0, S_BRAND_MANUFACTURER_ID, S_BRAND, S_BRAND_MANUFACTURER_ID},
    {0, 6, 0, 0, S_BRAND_NAME, S_BRAND, S_BRAND_NAME},
    {0, 0, 0, 0, S_CADR_ID, S_CUSTOMER_ADDRESS, S_CADR_ID},
    {0, 1, 0, 0, S_CADR_ADDRESS_STREET_NUMBER, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_STREET_NUMBER},
    {0, 1, 0, 0, S_CADR_ADDRESS_STREET_NAME1, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_STREET_NAME1},
    {0, 1, 0, 0, S_CADR_ADDRESS_STREET_NAME2, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_STREET_NAME2},
    {0, 1, 0, 0, S_CADR_ADDRESS_STREET_TYPE, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_STREET_TYPE},
    {0, 1, 0, 0, S_CADR_ADDRESS_SUITE_NUM, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_SUITE_NUM},
    {0, 1, 0, 0, S_CADR_ADDRESS_CITY, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_CITY},
    {0, 1, 0, 0, S_CADR_ADDRESS_COUNTY, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_COUNTY},
    {0, 1, 0, 0, S_CADR_ADDRESS_STATE, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_STATE},
    {0, 1, 0, 0, S_CADR_ADDRESS_ZIP, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_ZIP},
    {0, 1, 0, 0, S_CADR_ADDRESS_COUNTRY, S_CUSTOMER_ADDRESS, S_CADR_ADDRESS_COUNTRY},
    {0, 7, 0, 0, S_BADDR_ADDRESS, S_CUSTOMER_ADDRESS, CA_ADDRESS},
    {0, 1, 0, 0, S_CALL_CENTER_ID, S_CALL_CENTER, S_CALL_CENTER_ID},
    {0, 1, 0, 0, S_CALL_CENTER_DIVISION_ID, S_CALL_CENTER, S_CALL_CENTER_DIVISION_ID},
    {0, 1, 0, 0, S_CALL_CENTER_OPEN_DATE, S_CALL_CENTER, S_CALL_CENTER_OPEN_DATE},
    {0, 0, 0, 0, S_CALL_CENTER_CLOSED_DATE, S_CALL_CENTER, S_CALL_CENTER_CLOSED_DATE},
    {0, 0, 0, 0, S_CALL_CENTER_NAME, S_CALL_CENTER, S_CALL_CENTER_NAME},
    {0, 0, 0, 0, S_CALL_CENTER_CLASS, S_CALL_CENTER, S_CALL_CENTER_CLASS},
    {0, 1, 0, 0, S_CALL_CENTER_EMPLOYEES, S_CALL_CENTER, S_CALL_CENTER_EMPLOYEES},
    {0, 1, 0, 0, S_CALL_CENTER_SQFT, S_CALL_CENTER, S_CALL_CENTER_SQFT},
    {0, 1, 0, 0, S_CALL_CENTER_HOURS, S_CALL_CENTER, S_CALL_CENTER_HOURS},
    {0, 1, 0, 0, S_CALL_CENTER_MANAGER_ID, S_CALL_CENTER, S_CALL_CENTER_MANAGER_ID},
    {0, 1, 0, 0, S_CALL_CENTER_MARKET_ID, S_CALL_CENTER, S_CALL_CENTER_MARKET_ID},
    {0, 1, 0, 0, S_CALL_CENTER_ADDRESS_ID, S_CALL_CENTER, S_CALL_CENTER_ADDRESS_ID},
    {0, 1, 0, 0, S_CALL_CENTER_TAX_PERCENTAGE, S_CALL_CENTER, S_CALL_CENTER_TAX_PERCENTAGE},
    {0, 1, 0, 0, S_CALL_CENTER_SCD, S_CALL_CENTER, S_CALL_CENTER_SCD},
    {0, 0, 0, 0, S_CATALOG_NUMBER, S_CATALOG, S_CATALOG_NUMBER},
    {0, 1, 0, 0, S_CATALOG_START_DATE, S_CATALOG, S_CATALOG_START_DATE},
    {0, 1, 0, 0, S_CATALOG_END_DATE, S_CATALOG, S_CATALOG_END_DATE},
    {0, 10, 0, 0, S_CATALOG_DESC, S_CATALOG, S_CATALOG_DESC},
    {0, 1, 0, 0, S_CATALOG_TYPE, S_CATALOG, S_CATALOG_TYPE},
    {0, 0, 0, 0, S_CORD_ID, S_CATALOG_ORDER, S_CORD_ID},
    {0, 1, 0, 0, S_CORD_BILL_CUSTOMER_ID, S_CATALOG_ORDER, S_CORD_BILL_CUSTOMER_ID},
    {0, 2, 0, 0, S_CORD_SHIP_CUSTOMER_ID, S_CATALOG_ORDER, S_CORD_SHIP_CUSTOMER_ID},
    {0, 1, 0, 0, S_CORD_ORDER_DATE, S_CATALOG_ORDER, S_CORD_ORDER_DATE},
    {0, 1, 0, 0, S_CORD_ORDER_TIME, S_CATALOG_ORDER, S_CORD_ORDER_TIME},
    {0, 1, 0, 0, S_CORD_SHIP_MODE_ID, S_CATALOG_ORDER, S_CORD_SHIP_MODE_ID},
    {0, 1, 0, 0, S_CORD_CALL_CENTER_ID, S_CATALOG_ORDER, S_CORD_CALL_CENTER_ID},
    {0, 1, 0, 0, S_CLIN_ITEM_ID, S_CATALOG_ORDER, S_CLIN_ITEM_ID},
    {0, 100, 0, 0, S_CORD_COMMENT, S_CATALOG_ORDER, S_CORD_COMMENT},
    {0, 1, 0, 0, S_CLIN_ORDER_ID, S_CATALOG_ORDER_LINEITEM, S_CLIN_ORDER_ID},
    {0, 0, 0, 0, S_CLIN_LINE_NUMBER, S_CATALOG_ORDER_LINEITEM, S_CLIN_LINE_NUMBER},
    {0, 1, 0, 0, S_CLIN_PROMOTION_ID, S_CATALOG_ORDER_LINEITEM, S_CLIN_PROMOTION_ID},
    {0, 1, 0, 0, S_CLIN_QUANTITY, S_CATALOG_ORDER_LINEITEM, S_CLIN_QUANTITY},
    {0, 1, 0, 0, S_CLIN_COUPON_AMT, S_CATALOG_ORDER_LINEITEM, S_CLIN_COUPON_AMT},
    {0, 1, 0, 0, S_CLIN_WAREHOUSE_ID, S_CATALOG_ORDER_LINEITEM, S_CLIN_WAREHOUSE_ID},
    {0, 1, 0, 0, S_CLIN_SHIP_DATE, S_CATALOG_ORDER_LINEITEM, S_CLIN_SHIP_DATE},
    {0, 1, 0, 0, S_CLIN_CATALOG_ID, S_CATALOG_ORDER_LINEITEM, S_CLIN_CATALOG_ID},
    {0, 2, 0, 0, S_CLIN_CATALOG_PAGE_ID, S_CATALOG_ORDER_LINEITEM, S_CLIN_CATALOG_PAGE_ID},
    {0, 8, 0, 0, S_CLIN_PRICING, S_CATALOG_ORDER_LINEITEM, S_CLIN_PRICING},
    {0, 0, 0, 0, S_CLIN_SHIP_COST, S_CATALOG_ORDER_LINEITEM, S_CLIN_SHIP_COST},
    {0, 1, 0, 0, S_CLIN_IS_RETURNED, S_CATALOG_ORDER_LINEITEM, S_CLIN_IS_RETURNED},
    {0, 0, 0, 0, S_CLIN_PERMUTE, S_CATALOG_ORDER_LINEITEM, S_CLIN_PERMUTE},
    {0, 0, 0, 0, S_CATALOG_PAGE_CATALOG_NUMBER, S_CATALOG_PAGE, S_CATALOG_PAGE_CATALOG_NUMBER},
    {0, 0, 0, 0, S_CATALOG_PAGE_NUMBER, S_CATALOG_PAGE, S_CATALOG_PAGE_NUMBER},
    {0, 1, 0, 0, S_CATALOG_PAGE_DEPARTMENT, S_CATALOG_PAGE, S_CATALOG_PAGE_DEPARTMENT},
    {0, 1, 0, 0, S_CP_ID, S_CATALOG_PAGE, S_CP_ID},
    {0, 1, 0, 0, S_CP_START_DATE, S_CATALOG_PAGE, S_CP_START_DATE},
    {0, 1, 0, 0, S_CP_END_DATE, S_CATALOG_PAGE, S_CP_END_DATE},
    {0, 100, 0, 0, S_CP_DESCRIPTION, S_CATALOG_PAGE, CP_DESCRIPTION},
    {0, 1, 0, 0, S_CP_TYPE, S_CATALOG_PAGE, S_CP_TYPE},
    {0, 1, 0, 0, S_CATALOG_PROMOTIONAL_ITEM_CATALOG_NUMBER, S_CATALOG_PROMOTIONAL_ITEM,
     S_CATALOG_PROMOTIONAL_ITEM_CATALOG_NUMBER},
    {0, 1, 0, 0, S_CATALOG_PROMOTIONAL_ITEM_CATALOG_PAGE_NUMBER, S_CATALOG_PROMOTIONAL_ITEM,
     S_CATALOG_PROMOTIONAL_ITEM_CATALOG_PAGE_NUMBER},
    {0, 1, 0, 0, S_CATALOG_PROMOTIONAL_ITEM_ITEM_ID, S_CATALOG_PROMOTIONAL_ITEM, S_CATALOG_PROMOTIONAL_ITEM_ITEM_ID},
    {0, 0, 0, 0, S_CATALOG_PROMOTIONAL_ITEM_PROMOTION_ID, S_CATALOG_PROMOTIONAL_ITEM,
     S_CATALOG_PROMOTIONAL_ITEM_PROMOTION_ID},
    {0, 9, 0, 0, S_CRET_CALL_CENTER_ID, S_CATALOG_RETURNS, S_CRET_CALL_CENTER_ID},
    {0, 0, 0, 0, S_CRET_ORDER_ID, S_CATALOG_RETURNS, S_CRET_ORDER_ID},
    {0, 0, 0, 0, S_CRET_LINE_NUMBER, S_CATALOG_RETURNS, S_CRET_LINE_NUMBER},
    {0, 0, 0, 0, S_CRET_ITEM_ID, S_CATALOG_RETURNS, S_CRET_ITEM_ID},
    {0, 0, 0, 0, S_CRET_RETURN_CUSTOMER_ID, S_CATALOG_RETURNS, S_CRET_RETURN_CUSTOMER_ID},
    {0, 0, 0, 0, S_CRET_REFUND_CUSTOMER_ID, S_CATALOG_RETURNS, S_CRET_REFUND_CUSTOMER_ID},
    {0, 9, 0, 0, S_CRET_DATE, S_CATALOG_RETURNS, S_CRET_DATE},
    {0, 18, 0, 0, S_CRET_TIME, S_CATALOG_RETURNS, S_CRET_TIME},
    {0, 0, 0, 0, S_CRET_QUANTITY, S_CATALOG_RETURNS, S_CRET_QUANTITY},
    {0, 0, 0, 0, S_CRET_AMOUNT, S_CATALOG_RETURNS, S_CRET_AMOUNT},
    {0, 0, 0, 0, S_CRET_TAX, S_CATALOG_RETURNS, S_CRET_TAX},
    {0, 0, 0, 0, S_CRET_FEE, S_CATALOG_RETURNS, S_CRET_FEE},
    {0, 0, 0, 0, S_CRET_SHIP_COST, S_CATALOG_RETURNS, S_CRET_SHIP_COST},
    {0, 0, 0, 0, S_CRET_REFUNDED_CASH, S_CATALOG_RETURNS, S_CRET_REFUNDED_CASH},
    {0, 0, 0, 0, S_CRET_REVERSED_CHARGE, S_CATALOG_RETURNS, S_CRET_REVERSED_CHARGE},
    {0, 0, 0, 0, S_CRET_MERCHANT_CREDIT, S_CATALOG_RETURNS, S_CRET_MERCHANT_CREDIT},
    {0, 9, 0, 0, S_CRET_REASON_ID, S_CATALOG_RETURNS, S_CRET_REASON_ID},
    {0, 72, 0, 0, S_CRET_PRICING, S_CATALOG_RETURNS, S_CRET_PRICING},
    {0, 9, 0, 0, S_CRET_SHIPMODE_ID, S_CATALOG_RETURNS, S_CRET_SHIPMODE_ID},
    {0, 9, 0, 0, S_CRET_WAREHOUSE_ID, S_CATALOG_RETURNS, S_CRET_WAREHOUSE_ID},
    {0, 0, 0, 0, S_CRET_CATALOG_PAGE_ID, S_CATALOG_RETURNS, S_CRET_CATALOG_PAGE_ID},
    {0, 0, 0, 0, S_CATEGORY_ID, S_CATEGORY, S_CATEGORY_ID},
    {0, 0, 0, 0, S_CATEGORY_NAME, S_CATEGORY, S_CATEGORY_NAME},
    {0, 10, 0, 0, S_CATEGORY_DESC, S_CATEGORY, S_CATEGORY_DESC},
    {0, 0, 0, 0, S_CLASS_ID, S_CLASS, S_CLASS_ID},
    {0, 1, 0, 0, S_CLASS_SUBCAT_ID, S_CLASS, S_CLASS_SUBCAT_ID},
    {0, 10, 0, 0, S_CLASS_DESC, S_CLASS, S_CLASS_DESC},
    {0, 0, 0, 0, S_COMPANY_ID, S_COMPANY, S_COMPANY_ID},
    {0, 0, 0, 0, S_COMPANY_NAME, S_COMPANY, S_COMPANY_NAME},
    {0, 0, 0, 0, S_CUST_ID, S_CUSTOMER, S_CUST_ID},
    {0, 1, 0, 0, S_CUST_SALUTATION, S_CUSTOMER, S_CUST_SALUTATION},
    {0, 1, 0, 0, S_CUST_LAST_NAME, S_CUSTOMER, S_CUST_LAST_NAME},
    {0, 1, 0, 0, S_CUST_FIRST_NAME, S_CUSTOMER, S_CUST_FIRST_NAME},
    {0, 1, 0, 0, S_CUST_PREFERRED_FLAG, S_CUSTOMER, S_CUST_PREFERRED_FLAG},
    {0, 1, 0, 0, S_CUST_BIRTH_DATE, S_CUSTOMER, S_CUST_BIRTH_DATE},
    {0, 1, 0, 0, S_CUST_FIRST_PURCHASE_DATE, S_CUSTOMER, S_CUST_FIRST_PURCHASE_DATE},
    {0, 1, 0, 0, S_CUST_FIRST_SHIPTO_DATE, S_CUSTOMER, S_CUST_FIRST_SHIPTO_DATE},
    {0, 1, 0, 0, S_CUST_BIRTH_COUNTRY, S_CUSTOMER, S_CUST_BIRTH_COUNTRY},
    {0, 25, 0, 0, S_CUST_LOGIN, S_CUSTOMER, S_CUST_LOGIN},
    {0, 23, 0, 0, S_CUST_EMAIL, S_CUSTOMER, S_CUST_EMAIL},
    {0, 1, 0, 0, S_CUST_LAST_LOGIN, S_CUSTOMER, S_CUST_LAST_LOGIN},
    {0, 1, 0, 0, S_CUST_LAST_REVIEW, S_CUSTOMER, S_CUST_LAST_REVIEW},
    {0, 4, 0, 0, S_CUST_PRIMARY_MACHINE, S_CUSTOMER, S_CUST_PRIMARY_MACHINE},
    {0, 4, 0, 0, S_CUST_SECONDARY_MACHINE, S_CUSTOMER, S_CUST_SECONDARY_MACHINE},
    {0, 7, 0, 0, S_CUST_ADDRESS, S_CUSTOMER, S_CUST_ADDRESS},
    {0, 1, 0, 0, S_CUST_ADDRESS_STREET_NUM, S_CUSTOMER, S_CUST_ADDRESS_STREET_NUM},
    {0, 1, 0, 0, S_CUST_ADDRESS_STREET_NAME1, S_CUSTOMER, S_CUST_ADDRESS_STREET_NAME1},
    {0, 1, 0, 0, S_CUST_ADDRESS_STREET_NAME2, S_CUSTOMER, S_CUST_ADDRESS_STREET_NAME2},
    {0, 1, 0, 0, S_CUST_ADDRESS_STREET_TYPE, S_CUSTOMER, S_CUST_ADDRESS_STREET_TYPE},
    {0, 1, 0, 0, S_CUST_ADDRESS_SUITE_NUM, S_CUSTOMER, S_CUST_ADDRESS_SUITE_NUM},
    {0, 1, 0, 0, S_CUST_ADDRESS_CITY, S_CUSTOMER, S_CUST_ADDRESS_CITY},
    {0, 1, 0, 0, S_CUST_ADDRESS_ZIP, S_CUSTOMER, S_CUST_ADDRESS_ZIP},
    {0, 1, 0, 0, S_CUST_ADDRESS_COUNTY, S_CUSTOMER, S_CUST_ADDRESS_COUNTY},
    {0, 1, 0, 0, S_CUST_ADDRESS_STATE, S_CUSTOMER, S_CUST_ADDRESS_STATE},
    {0, 1, 0, 0, S_CUST_ADDRESS_COUNTRY, S_CUSTOMER, S_CUST_ADDRESS_COUNTRY},
    {0, 1, 0, 0, S_CUST_LOCATION_TYPE, S_CUSTOMER, S_CUST_LOCATION_TYPE},
    {0, 1, 0, 0, S_CUST_GENDER, S_CUSTOMER, S_CUST_GENDER},
    {0, 1, 0, 0, S_CUST_MARITAL_STATUS, S_CUSTOMER, S_CUST_MARITAL_STATUS},
    {0, 1, 0, 0, S_CUST_EDUCATION, S_CUSTOMER, S_CUST_EDUCATION},
    {0, 1, 0, 0, S_CUST_CREDIT_RATING, S_CUSTOMER, S_CUST_CREDIT_RATING},
    {0, 1, 0, 0, S_CUST_PURCHASE_ESTIMATE, S_CUSTOMER, S_CUST_PURCHASE_ESTIMATE},
    {0, 1, 0, 0, S_CUST_BUY_POTENTIAL, S_CUSTOMER, S_CUST_BUY_POTENTIAL},
    {0, 1, 0, 0, S_CUST_DEPENDENT_CNT, S_CUSTOMER, S_CUST_DEPENDENT_CNT},
    {0, 1, 0, 0, S_CUST_EMPLOYED_CNT, S_CUSTOMER, S_CUST_EMPLOYED_CNT},
    {0, 1, 0, 0, S_CUST_COLLEGE_CNT, S_CUSTOMER, S_CUST_COLLEGE_CNT},
    {0, 1, 0, 0, S_CUST_VEHICLE_CNT, S_CUSTOMER, S_CUST_VEHICLE_CNT},
    {0, 1, 0, 0, S_CUST_INCOME, S_CUSTOMER, S_CUST_INCOME},
    {0, 0, 0, 0, S_DIVISION_ID, S_DIVISION, S_DIVISION_ID},
    {0, 0, 0, 0, S_DIVISION_COMPANY, S_DIVISION, S_DIVISION_COMPANY},
    {0, 0, 0, 0, S_DIVISION_NAME, S_DIVISION, S_DIVISION_NAME},
    {0, 1, 0, 0, S_INVN_WAREHOUSE, S_INVENTORY, S_INVN_WAREHOUSE},
    {0, 1, 0, 0, S_INVN_ITEM, S_INVENTORY, S_INVN_ITEM},
    {0, 2, 0, 0, S_INVN_DATE, S_INVENTORY, S_INVN_DATE},
    {0, 1, 0, 0, S_INVN_QUANTITY, S_INVENTORY, S_INVN_QUANTITY},
    {0, 1, 0, 0, S_ITEM_ID, S_ITEM, S_ITEM_ID},
    {0, 0, 0, 0, S_ITEM_PERMUTE, S_ITEM, S_ITEM_PERMUTE},
    {0, 1, 0, 0, S_ITEM_PRODUCT_ID, S_ITEM, S_ITEM_PRODUCT_ID},
    {0, 200, 0, 0, S_ITEM_DESC, S_ITEM, I_ITEM_DESC},
    {0, 2, 0, 0, S_ITEM_LIST_PRICE, S_ITEM, I_CURRENT_PRICE},
    {0, 1, 0, 0, S_ITEM_WHOLESALE_COST, S_ITEM, I_WHOLESALE_COST},
    {0, 1, 0, 0, S_ITEM_MANAGER_ID, S_ITEM, I_MANAGER_ID},
    {0, 1, 0, 0, S_ITEM_SIZE, S_ITEM, I_SIZE},
    {0, 50, 0, 0, S_ITEM_FORMULATION, S_ITEM, I_FORMULATION},
    {0, 1, 0, 0, S_ITEM_FLAVOR, S_ITEM, I_COLOR},
    {0, 1, 0, 0, S_ITEM_UNITS, S_ITEM, I_UNITS},
    {0, 1, 0, 0, S_ITEM_CONTAINER, S_ITEM, I_CONTAINER},
    {0, 1, 0, 0, S_ITEM_SCD, S_ITEM, I_SCD},
    {0, 0, 0, 0, S_MANAGER_ID, S_MANAGER, S_MANAGER_ID},
    {0, 2, 0, 0, S_MANAGER_NAME, S_MANAGER, S_MANAGER_NAME},
    {0, 0, 0, 0, S_MANUFACTURER_ID, S_MANUFACTURER, S_MANUFACTURER_ID},
    {0, 0, 0, 0, S_MANUFACTURER_NAME, S_MANUFACTURER, S_MANUFACTURER_NAME},
    {0, 0, 0, 0, S_MARKET_ID, S_MARKET, S_MARKET_ID},
    {0, 0, 0, 0, S_MARKET_CLASS_NAME, S_MARKET, S_MARKET_CLASS_NAME},
    {0, 10, 0, 0, S_MARKET_DESC, S_MARKET, S_MARKET_DESC},
    {0, 1, 0, 0, S_MARKET_MANAGER_ID, S_MARKET, S_MARKET_MANAGER_ID},
    {0, 0, 0, 0, S_PRODUCT_ID, S_PRODUCT, S_PRODUCT_ID},
    {0, 1, 0, 0, S_PRODUCT_BRAND_ID, S_PRODUCT, S_PRODUCT_BRAND_ID},
    {0, 0, 0, 0, S_PRODUCT_NAME, S_PRODUCT, S_PRODUCT_NAME},
    {0, 1, 0, 0, S_PRODUCT_TYPE, S_PRODUCT, S_PRODUCT_TYPE},
    {0, 1, 0, 0, S_PROMOTION_ID, S_PROMOTION, S_PROMOTION_ID},
    {0, 1, 0, 0, S_PROMOTION_ITEM_ID, S_PROMOTION, S_PROMOTION_ITEM_ID},
    {0, 1, 0, 0, S_PROMOTION_START_DATE, S_PROMOTION, S_PROMOTION_START_DATE},
    {0, 1, 0, 0, S_PROMOTION_END_DATE, S_PROMOTION, S_PROMOTION_END_DATE},
    {0, 1, 0, 0, S_PROMOTION_COST, S_PROMOTION, S_PROMOTION_COST},
    {0, 1, 0, 0, S_PROMOTION_RESPONSE_TARGET, S_PROMOTION, S_PROMOTION_RESPONSE_TARGET},
    {0, 0, 0, 0, S_PROMOTION_DMAIL, S_PROMOTION, S_PROMOTION_DMAIL},
    {0, 0, 0, 0, S_PROMOTION_EMAIL, S_PROMOTION, S_PROMOTION_EMAIL},
    {0, 0, 0, 0, S_PROMOTION_CATALOG, S_PROMOTION, S_PROMOTION_CATALOG},
    {0, 0, 0, 0, S_PROMOTION_TV, S_PROMOTION, S_PROMOTION_TV},
    {0, 0, 0, 0, S_PROMOTION_RADIO, S_PROMOTION, S_PROMOTION_RADIO},
    {0, 0, 0, 0, S_PROMOTION_PRESS, S_PROMOTION, S_PROMOTION_PRESS},
    {0, 0, 0, 0, S_PROMOTION_EVENT, S_PROMOTION, S_PROMOTION_EVENT},
    {0, 0, 0, 0, S_PROMOTION_DEMO, S_PROMOTION, S_PROMOTION_DEMO},
    {0, 100, 0, 0, S_PROMOTION_DETAILS, S_PROMOTION, P_CHANNEL_DETAILS},
    {0, 1, 0, 0, S_PROMOTION_PURPOSE, S_PROMOTION, S_PROMOTION_PURPOSE},
    {0, 1, 0, 0, S_PROMOTION_DISCOUNT_ACTIVE, S_PROMOTION, S_PROMOTION_DISCOUNT_ACTIVE},
    {0, 1, 0, 0, S_PROMOTION_DISCOUNT_PCT, S_PROMOTION, S_PROMOTION_DISCOUNT_PCT},
    {0, 0, 0, 0, S_PROMOTION_NAME, S_PROMOTION, S_PROMOTION_NAME},
    {0, 1, 0, 0, S_PROMOTION_BITFIELD, S_PROMOTION, S_PROMOTION_BITFIELD},
    {0, 0, 0, 0, S_PURCHASE_ID, S_PURCHASE, S_PURCHASE_ID},
    {0, 1, 0, 0, S_PURCHASE_STORE_ID, S_PURCHASE, S_PURCHASE_STORE_ID},
    {0, 1, 0, 0, S_PURCHASE_CUSTOMER_ID, S_PURCHASE, S_PURCHASE_CUSTOMER_ID},
    {0, 1, 0, 0, S_PURCHASE_DATE, S_PURCHASE, S_PURCHASE_DATE},
    {0, 1, 0, 0, S_PURCHASE_TIME, S_PURCHASE, S_PURCHASE_TIME},
    {0, 1, 0, 0, S_PURCHASE_REGISTER, S_PURCHASE, S_PURCHASE_REGISTER},
    {0, 1, 0, 0, S_PURCHASE_CLERK, S_PURCHASE, S_PURCHASE_CLERK},
    {0, 100, 0, 0, S_PURCHASE_COMMENT, S_PURCHASE, S_PURCHASE_COMMENT},
    {0, 7, 0, 0, S_PURCHASE_PRICING, S_PURCHASE, S_PURCHASE_PRICING},
    {0, 1, 0, 0, S_PLINE_ITEM_ID, S_PURCHASE, S_PLINE_ITEM_ID},
    {0, 12, 0, 0, S_PLINE_PURCHASE_ID, S_PURCHASE_LINEITEM, S_PLINE_PURCHASE_ID},
    {0, 12, 0, 0, S_PLINE_NUMBER, S_PURCHASE_LINEITEM, S_PLINE_NUMBER},
    {0, 12, 0, 0, S_PLINE_PROMOTION_ID, S_PURCHASE_LINEITEM, S_PLINE_PROMOTION_ID},
    {0, 12, 0, 0, S_PLINE_SALE_PRICE, S_PURCHASE_LINEITEM, S_PLINE_SALE_PRICE},
    {0, 12, 0, 0, S_PLINE_QUANTITY, S_PURCHASE_LINEITEM, S_PLINE_QUANTITY},
    {0, 12, 0, 0, S_PLINE_COUPON_AMT, S_PURCHASE_LINEITEM, S_PLINE_COUPON_AMT},
    {0, 1200, 0, 0, S_PLINE_COMMENT, S_PURCHASE_LINEITEM, S_PLINE_COMMENT},
    {0, 96, 0, 0, S_PLINE_PRICING, S_PURCHASE_LINEITEM, S_PLINE_PRICING},
    {0, 12, 0, 0, S_PLINE_IS_RETURNED, S_PURCHASE_LINEITEM, S_PLINE_IS_RETURNED},
    {0, 0, 0, 0, S_PLINE_PERMUTE, S_PURCHASE_LINEITEM, S_PLINE_PERMUTE},
    {0, 0, 0, 0, S_REASON_ID, S_REASON, S_REASON_ID},
    {0, 10, 0, 0, S_REASON_DESC, S_REASON, S_REASON_DESC},
    {0, 1, 0, 0, S_STORE_ID, S_STORE, S_STORE_ID},
    {0, 1, 0, 0, S_STORE_ADDRESS_ID, S_STORE, S_STORE_ADDRESS_ID},
    {0, 1, 0, 0, S_STORE_DIVISION_ID, S_STORE, S_STORE_DIVISION_ID},
    {0, 1, 0, 0, S_STORE_OPEN_DATE, S_STORE, S_STORE_OPEN_DATE},
    {0, 1, 0, 0, S_STORE_CLOSE_DATE, S_STORE, S_STORE_CLOSE_DATE},
    {0, 0, 0, 0, S_STORE_NAME, S_STORE, S_STORE_NAME},
    {0, 0, 0, 0, S_STORE_CLASS, S_STORE, S_STORE_CLASS},
    {0, 1, 0, 0, S_STORE_EMPLOYEES, S_STORE, S_STORE_EMPLOYEES},
    {0, 1, 0, 0, S_STORE_FLOOR_SPACE, S_STORE, S_STORE_FLOOR_SPACE},
    {0, 1, 0, 0, S_STORE_HOURS, S_STORE, S_STORE_HOURS},
    {0, 0, 0, 0, S_STORE_MARKET_MANAGER_ID, S_STORE, S_STORE_MARKET_MANAGER_ID},
    {0, 1, 0, 0, S_STORE_MANAGER_ID, S_STORE, S_STORE_MANAGER_ID},
    {0, 1, 0, 0, S_STORE_MARKET_ID, S_STORE, S_STORE_MARKET_ID},
    {0, 1, 0, 0, S_STORE_GEOGRAPHY_CLASS, S_STORE, S_STORE_GEOGRAPHY_CLASS},
    {0, 1, 0, 0, S_STORE_TAX_PERCENTAGE, S_STORE, S_STORE_TAX_PERCENTAGE},
    {0, 1, 0, 0, S_SITM_PROMOTION_ID, S_STORE_PROMOTIONAL_ITEM, S_SITM_PROMOTION_ID},
    {0, 1, 0, 0, S_SITM_ITEM_ID, S_STORE_PROMOTIONAL_ITEM, S_SITM_ITEM_ID},
    {0, 1, 0, 0, S_SITM_STORE_ID, S_STORE_PROMOTIONAL_ITEM, S_SITM_STORE_ID},
    {0, 0, 0, 0, S_SRET_STORE_ID, S_STORE_RETURNS, S_SRET_STORE_ID},
    {0, 0, 0, 0, S_SRET_PURCHASE_ID, S_STORE_RETURNS, S_SRET_PURCHASE_ID},
    {0, 0, 0, 0, S_SRET_LINENUMBER, S_STORE_RETURNS, S_SRET_LINENUMBER},
    {0, 0, 0, 0, S_SRET_ITEM_ID, S_STORE_RETURNS, S_SRET_ITEM_ID},
    {0, 0, 0, 0, S_SRET_CUSTOMER_ID, S_STORE_RETURNS, S_SRET_CUSTOMER_ID},
    {0, 24, 0, 0, S_SRET_RETURN_DATE, S_STORE_RETURNS, S_SRET_RETURN_DATE},
    {0, 12, 0, 0, S_SRET_RETURN_TIME, S_STORE_RETURNS, S_SRET_RETURN_TIME},
    {0, 0, 0, 0, S_SRET_TICKET_NUMBER, S_STORE_RETURNS, S_SRET_TICKET_NUMBER},
    {0, 0, 0, 0, S_SRET_RETURN_QUANTITY, S_STORE_RETURNS, S_SRET_RETURN_QUANTITY},
    {0, 0, 0, 0, S_SRET_RETURN_AMT, S_STORE_RETURNS, S_SRET_RETURN_AMT},
    {0, 0, 0, 0, S_SRET_RETURN_TAX, S_STORE_RETURNS, S_SRET_RETURN_TAX},
    {0, 0, 0, 0, S_SRET_RETURN_FEE, S_STORE_RETURNS, S_SRET_RETURN_FEE},
    {0, 0, 0, 0, S_SRET_RETURN_SHIP_COST, S_STORE_RETURNS, S_SRET_RETURN_SHIP_COST},
    {0, 0, 0, 0, S_SRET_REFUNDED_CASH, S_STORE_RETURNS, S_SRET_REFUNDED_CASH},
    {0, 0, 0, 0, S_SRET_REVERSED_CHARGE, S_STORE_RETURNS, S_SRET_REVERSED_CHARGE},
    {0, 0, 0, 0, S_SRET_MERCHANT_CREDIT, S_STORE_RETURNS, S_SRET_MERCHANT_CREDIT},
    {0, 12, 0, 0, S_SRET_REASON_ID, S_STORE_RETURNS, S_SRET_REASON_ID},
    {0, 84, 0, 0, S_SRET_PRICING, S_STORE_RETURNS, S_SRET_PRICING},
    {0, 0, 0, 0, S_SBCT_ID, S_SUBCATEGORY, S_SBCT_ID},
    {0, 1, 0, 0, S_SBCT_CATEGORY_ID, S_SUBCATEGORY, S_SBCT_CATEGORY_ID},
    {0, 0, 0, 0, S_SBCT_NAME, S_SUBCATEGORY, S_SBCT_NAME},
    {0, 10, 0, 0, S_SBCT_DESC, S_SUBCATEGORY, S_SBCT_DESC},
    {0, 0, 0, 0, S_SUBC_ID, S_SUBCLASS, S_SUBC_ID},
    {0, 1, 0, 0, S_SUBC_CLASS_ID, S_SUBCLASS, S_SUBC_CLASS_ID},
    {0, 0, 0, 0, S_SUBC_NAME, S_SUBCLASS, S_SUBC_NAME},
    {0, 10, 0, 0, S_SUBC_DESC, S_SUBCLASS, S_SUBC_DESC},
    {0, 1, 0, 0, S_WRHS_ID, S_WAREHOUSE, S_WRHS_ID},
    {0, 10, 0, 0, S_WRHS_DESC, S_WAREHOUSE, S_WRHS_DESC},
    {0, 1, 0, 0, S_WRHS_SQFT, S_WAREHOUSE, S_WRHS_SQFT},
    {0, 1, 0, 0, S_WRHS_ADDRESS_ID, S_WAREHOUSE, S_WRHS_ADDRESS_ID},
    {0, 1, 0, 0, S_WORD_ID, S_WEB_ORDER, S_WORD_ID},
    {0, 1, 0, 0, S_WORD_BILL_CUSTOMER_ID, S_WEB_ORDER, S_WORD_BILL_CUSTOMER_ID},
    {0, 2, 0, 0, S_WORD_SHIP_CUSTOMER_ID, S_WEB_ORDER, S_WORD_SHIP_CUSTOMER_ID},
    {0, 1, 0, 0, S_WORD_ORDER_DATE, S_WEB_ORDER, S_WORD_ORDER_DATE},
    {0, 1, 0, 0, S_WORD_ORDER_TIME, S_WEB_ORDER, S_WORD_ORDER_TIME},
    {0, 1, 0, 0, S_WORD_SHIP_MODE_ID, S_WEB_ORDER, S_WORD_SHIP_MODE_ID},
    {0, 1, 0, 0, S_WORD_WEB_SITE_ID, S_WEB_ORDER, S_WORD_WEB_SITE_ID},
    {0, 100, 0, 0, S_WORD_COMMENT, S_WEB_ORDER, S_WORD_COMMENT},
    {0, 1, 0, 0, S_WLIN_ITEM_ID, S_WEB_ORDER, S_WLIN_ITEM_ID},
    {0, 12, 0, 0, S_WLIN_ID, S_WEB_ORDER_LINEITEM, S_WLIN_ID},
    {0, 0, 0, 0, S_WLIN_LINE_NUMBER, S_WEB_ORDER_LINEITEM, S_WLIN_LINE_NUMBER},
    {0, 12, 0, 0, S_WLIN_PROMOTION_ID, S_WEB_ORDER_LINEITEM, S_WLIN_PROMOTION_ID},
    {0, 12, 0, 0, S_WLIN_QUANTITY, S_WEB_ORDER_LINEITEM, S_WLIN_QUANTITY},
    {0, 12, 0, 0, S_WLIN_COUPON_AMT, S_WEB_ORDER_LINEITEM, S_WLIN_COUPON_AMT},
    {0, 12, 0, 0, S_WLIN_WAREHOUSE_ID, S_WEB_ORDER_LINEITEM, S_WLIN_WAREHOUSE_ID},
    {0, 12, 0, 0, S_WLIN_SHIP_DATE, S_WEB_ORDER_LINEITEM, S_WLIN_SHIP_DATE},
    {0, 12, 0, 0, S_WLIN_WEB_PAGE_ID, S_WEB_ORDER_LINEITEM, S_WLIN_WEB_PAGE_ID},
    {0, 96, 0, 0, S_WLIN_PRICING, S_WEB_ORDER_LINEITEM, S_WLIN_PRICING},
    {0, 0, 0, 0, S_WLIN_SHIP_COST, S_WEB_ORDER_LINEITEM, S_WLIN_SHIP_COST},
    {0, 12, 0, 0, S_WLIN_IS_RETURNED, S_WEB_ORDER_LINEITEM, S_WLIN_IS_RETURNED},
    {0, 0, 0, 0, S_WLIN_PERMUTE, S_WEB_ORDER_LINEITEM, S_WLIN_PERMUTE},
    {0, 1, 0, 0, S_WPAG_SITE_ID, S_WEB_PAGE, S_WPAG_SITE_ID},
    {0, 1, 0, 0, S_WPAG_ID, S_WEB_PAGE, S_WPAG_ID},
    {0, 1, 0, 0, S_WPAG_CREATE_DATE, S_WEB_PAGE, S_WPAG_CREATE_DATE},
    {0, 1, 0, 0, S_WPAG_ACCESS_DATE, S_WEB_PAGE, S_WPAG_ACCESS_DATE},
    {0, 1, 0, 0, S_WPAG_AUTOGEN_FLAG, S_WEB_PAGE, S_WPAG_AUTOGEN_FLAG},
    {0, 1, 0, 0, S_WPAG_DEPARTMENT, S_WEB_PAGE, S_WPAG_DEPARTMENT},
    {0, 1, 0, 0, S_WPAG_URL, S_WEB_PAGE, S_WPAG_URL},
    {0, 1, 0, 0, S_WPAG_TYPE, S_WEB_PAGE, S_WPAG_TYPE},
    {0, 1, 0, 0, S_WPAG_CHAR_CNT, S_WEB_PAGE, S_WPAG_CHAR_CNT},
    {0, 1, 0, 0, S_WPAG_LINK_CNT, S_WEB_PAGE, S_WPAG_LINK_CNT},
    {0, 1, 0, 0, S_WPAG_IMAGE_CNT, S_WEB_PAGE, S_WPAG_IMAGE_CNT},
    {0, 1, 0, 0, S_WPAG_MAX_AD_CNT, S_WEB_PAGE, S_WPAG_MAX_AD_CNT},
    {0, 0, 0, 0, S_WPAG_PERMUTE, S_WEB_PAGE, S_WPAG_PERMUTE},
    {0, 1, 0, 0, S_WITM_SITE_ID, S_WEB_PROMOTIONAL_ITEM, S_WITM_SITE_ID},
    {0, 1, 0, 0, S_WITM_PAGE_ID, S_WEB_PROMOTIONAL_ITEM, S_WITM_PAGE_ID},
    {0, 1, 0, 0, S_WITM_ITEM_ID, S_WEB_PROMOTIONAL_ITEM, S_WITM_ITEM_ID},
    {0, 1, 0, 0, S_WITM_PROMOTION_ID, S_WEB_PROMOTIONAL_ITEM, S_WITM_PROMOTION_ID},
    {0, 0, 0, 0, S_WRET_SITE_ID, S_WEB_RETURNS, S_WRET_SITE_ID},
    {0, 0, 0, 0, S_WRET_ORDER_ID, S_WEB_RETURNS, S_WRET_ORDER_ID},
    {0, 0, 0, 0, S_WRET_LINE_NUMBER, S_WEB_RETURNS, S_WRET_LINE_NUMBER},
    {0, 0, 0, 0, S_WRET_ITEM_ID, S_WEB_RETURNS, S_WRET_ITEM_ID},
    {0, 0, 0, 0, S_WRET_RETURN_CUST_ID, S_WEB_RETURNS, S_WRET_RETURN_CUST_ID},
    {0, 0, 0, 0, S_WRET_REFUND_CUST_ID, S_WEB_RETURNS, S_WRET_REFUND_CUST_ID},
    {0, 24, 0, 0, S_WRET_RETURN_DATE, S_WEB_RETURNS, S_WRET_RETURN_DATE},
    {0, 12, 0, 0, S_WRET_RETURN_TIME, S_WEB_RETURNS, S_WRET_RETURN_TIME},
    {0, 12, 0, 0, S_WRET_REASON_ID, S_WEB_RETURNS, S_WRET_REASON_ID},
    {0, 84, 0, 0, S_WRET_PRICING, S_WEB_RETURNS, S_WRET_PRICING},
    {0, 1, 0, 0, S_WSIT_ID, S_WEB_SITE, S_WSIT_ID},
    {0, 1, 0, 0, S_WSIT_OPEN_DATE, S_WEB_SITE, S_WSIT_OPEN_DATE},
    {0, 1, 0, 0, S_WSIT_CLOSE_DATE, S_WEB_SITE, S_WSIT_CLOSE_DATE},
    {0, 0, 0, 0, S_WSIT_NAME, S_WEB_SITE, S_WSIT_NAME},
    {0, 1, 0, 0, S_WSIT_ADDRESS_ID, S_WEB_SITE, S_WSIT_ADDRESS_ID},
    {0, 1, 0, 0, S_WSIT_DIVISION_ID, S_WEB_SITE, S_WSIT_DIVISION_ID},
    {0, 1, 0, 0, S_WSIT_CLASS, S_WEB_SITE, S_WSIT_CLASS},
    {0, 1, 0, 0, S_WSIT_MANAGER_ID, S_WEB_SITE, S_WSIT_MANAGER_ID},
    {0, 1, 0, 0, S_WSIT_MARKET_ID, S_WEB_SITE, S_WSIT_MARKET_ID},
    {0, 1, 0, 0, S_WSIT_TAX_PERCENTAGE, S_WEB_SITE, S_WSIT_TAX_PERCENTAGE},
    {0, 0, 0, 0, S_ZIPG_ZIP, S_ZIPG, S_ZIPG_ZIP},
    {0, 0, 0, 0, S_ZIPG_GMT, S_ZIPG, S_ZIPG_GMT},
    {-1, -1, -1, -1, -1, -1, -1}};
#endif
