//===----------------------------------------------------------------------===//
//                         DuckDB
//
// core_functions/scalar/debug_functions.hpp
//
//
//===----------------------------------------------------------------------===//
// This file is automatically generated by scripts/generate_functions.py
// Do not edit this file manually, your changes will be overwritten
//===----------------------------------------------------------------------===//

#pragma once

#include "duckdb/function/function_set.hpp"

namespace duckdb {

struct VectorTypeFun {
	static constexpr const char *Name = "vector_type";
	static constexpr const char *Parameters = "col";
	static constexpr const char *Description = "Returns the VectorType of a given column";
	static constexpr const char *Example = "vector_type(col)";
	static constexpr const char *Categories = "";

	static ScalarFunction GetFunction();
};

} // namespace duckdb
