//===----------------------------------------------------------------------===//
// This file is automatically generated by scripts/generate_serialization.py
// Do not edit this file manually, your changes will be overwritten
//===----------------------------------------------------------------------===//

#include "duckdb/common/serializer/serializer.hpp"
#include "duckdb/common/serializer/deserializer.hpp"
#include "json_transform.hpp"

namespace duckdb {

void JSONTransformOptions::Serialize(Serializer &serializer) const {
	serializer.WritePropertyWithDefault<bool>(100, "strict_cast", strict_cast);
	serializer.WritePropertyWithDefault<bool>(101, "error_duplicate_key", error_duplicate_key);
	serializer.WritePropertyWithDefault<bool>(102, "error_missing_key", error_missing_key);
	serializer.WritePropertyWithDefault<bool>(103, "error_unknown_key", error_unknown_key);
	serializer.WritePropertyWithDefault<bool>(104, "delay_error", delay_error);
}

JSONTransformOptions JSONTransformOptions::Deserialize(Deserializer &deserializer) {
	JSONTransformOptions result;
	deserializer.ReadPropertyWithDefault<bool>(100, "strict_cast", result.strict_cast);
	deserializer.ReadPropertyWithDefault<bool>(101, "error_duplicate_key", result.error_duplicate_key);
	deserializer.ReadPropertyWithDefault<bool>(102, "error_missing_key", result.error_missing_key);
	deserializer.ReadPropertyWithDefault<bool>(103, "error_unknown_key", result.error_unknown_key);
	deserializer.ReadPropertyWithDefault<bool>(104, "delay_error", result.delay_error);
	return result;
}

} // namespace duckdb
