/*
 * Legal Notice
 *
 * This document and associated source code (the "Work") is a part of a
 * benchmark specification maintained by the TPC.
 *
 * The TPC reserves all right, title, and interest to the Work as provided
 * under U.S. and international laws, including without limitation all patent
 * and trademark rights therein.
 *
 * No Warranty
 *
 * 1.1 TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THE INFORMATION
 *     CONTAINED HEREIN IS PROVIDED "AS IS" AND WITH ALL FAULTS, AND THE
 *     AUTHORS AND DEVELOPERS OF THE WORK HEREBY DISCLAIM ALL OTHER
 *     WARRANTIES AND CONDITIONS, EITHER EXPRESS, IMPLIED OR STATUTORY,
 *     INCLUDING, BUT NOT LIMITED TO, ANY (IF ANY) IMPLIED WARRANTIES,
 *     DUTIES OR CONDITIONS OF MERCHANTABILITY, OF FITNESS FOR A PARTICULAR
 *     PURPOSE, OF ACCURACY OR COMPLETENESS OF RESPONSES, OF RESULTS, OF
 *     WORKMANLIKE EFFORT, OF LACK OF VIRUSES, AND OF LACK OF NEGLIGENCE.
 *     ALSO, THERE IS NO WARRANTY OR CONDITION OF TITLE, QUIET ENJOYMENT,
 *     QUIET POSSESSION, CORRESPONDENCE TO DESCRIPTION OR NON-INFRINGEMENT
 *     WITH REGARD TO THE WORK.
 * 1.2 IN NO EVENT WILL ANY AUTHOR OR DEVELOPER OF THE WORK BE LIABLE TO
 *     ANY OTHER PARTY FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO THE
 *     COST OF PROCURING SUBSTITUTE GOODS OR SERVICES, LOST PROFITS, LOSS
 *     OF USE, LOSS OF DATA, OR ANY INCIDENTAL, CONSEQUENTIAL, DIRECT,
 *     INDIRECT, OR SPECIAL DAMAGES WHETHER UNDER CONTRACT, TORT, WARRANTY,
 *     OR OTHERWISE, ARISING IN ANY WAY OUT OF THIS OR ANY OTHER AGREEMENT
 *     RELATING TO THE WORK, WHETHER OR NOT SUCH AUTHOR OR DEVELOPER HAD
 *     ADVANCE NOTICE OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * Contributors
 * - Sergey Vasilevskiy
 */

/*
 *   Class for generation and load output to stdout.
 */

#ifndef EGEN_GENERATE_AND_LOAD_STANDARD_OUTPUT_H
#define EGEN_GENERATE_AND_LOAD_STANDARD_OUTPUT_H

#include <cstdio>
#include <string>
#include "EGenGenerateAndLoadBaseOutput.h"

using namespace std;

namespace TPCE {

class CGenerateAndLoadStandardOutput : public CGenerateAndLoadBaseOutput {
public:
	/*
	 *  Output beginning of table generation.
	 *
	 *  PARAMETERS:
	 *           IN  szMsg       - string to output to the user
	 *
	 *  RETURNS:
	 *           none.
	 */
	void OutputStart(string szMsg) {
		// printf("%s", szMsg.c_str());
		// fflush(stdout); // in case there is no newline character in szMsg
	}

	/*
	 *  Output progress of table generation.
	 *
	 *  PARAMETERS:
	 *           IN  szMsg       - string to output to the user
	 *
	 *  RETURNS:
	 *           none.
	 */
	void OutputProgress(string szMsg) {
		// printf("%s", szMsg.c_str());
		// fflush(stdout); // in case there is no newline character in szMsg
	}

	/*
	 *  Output completion of table generation.
	 *
	 *  PARAMETERS:
	 *           IN  szMsg       - string to output to the user
	 *
	 *  RETURNS:
	 *           none.
	 */
	void OutputComplete(string szMsg) {
		// printf("%s", szMsg.c_str());
		// fflush(stdout); // in case there is no newline character in szMsg
	}

	/*
	 *  Output end-of-line.
	 *
	 *  PARAMETERS:
	 *           IN  szMsg       - string to output to the user
	 *
	 *  RETURNS:
	 *           none.
	 */
	void OutputNewline() {
		// printf("\n");
	}
};

} // namespace TPCE

#endif // EGEN_GENERATE_AND_LOAD_STANDARD_OUTPUT_H
