create table web_returns(
	wr_returned_date_sk integer,
	wr_returned_time_sk integer,
	wr_item_sk integer not null,
	wr_refunded_customer_sk integer,
	wr_refunded_cdemo_sk integer,
	wr_refunded_hdemo_sk integer,
	wr_refunded_addr_sk integer,
	wr_returning_customer_sk integer,
	wr_returning_cdemo_sk integer,
	wr_returning_hdemo_sk integer,
	wr_returning_addr_sk integer,
	wr_web_page_sk integer,
	wr_reason_sk integer,
	wr_order_number integer not null,
	wr_return_quantity integer,
	wr_return_amt decimal(7,2),
	wr_return_tax decimal(7,2),
	wr_return_amt_inc_tax decimal(7,2),
	wr_fee decimal(7,2),
	wr_return_ship_cost decimal(7,2),
	wr_refunded_cash decimal(7,2),
	wr_reversed_charge decimal(7,2),
	wr_account_credit decimal(7,2),
	wr_net_loss decimal(7,2)
);
