/*
 * THIS IS A GENERATED FILE
 * SEE COLUMNS.LIST
 */
#ifndef COLUMNS_H
#define COLUMNS_H
#define CALL_CENTER_START                              1
#define CC_CALL_CENTER_SK                              1
#define CC_CALL_CENTER_ID                              2
#define CC_REC_START_DATE_ID                           3
#define CC_REC_END_DATE_ID                             4
#define CC_CLOSED_DATE_ID                              5
#define CC_OPEN_DATE_ID                                6
#define CC_NAME                                        7
#define CC_CLASS                                       8
#define CC_EMPLOYEES                                   9
#define CC_SQ_FT                                       10
#define CC_HOURS                                       11
#define CC_MANAGER                                     12
#define CC_MARKET_ID                                   13
#define CC_MARKET_CLASS                                14
#define CC_MARKET_DESC                                 15
#define CC_MARKET_MANAGER                              16
#define CC_DIVISION                                    17
#define CC_DIVISION_NAME                               18
#define CC_COMPANY                                     19
#define CC_COMPANY_NAME                                20
#define CC_STREET_NUMBER                               21
#define CC_STREET_NAME                                 22
#define CC_STREET_TYPE                                 23
#define CC_SUITE_NUMBER                                24
#define CC_CITY                                        25
#define CC_COUNTY                                      26
#define CC_STATE                                       27
#define CC_ZIP                                         28
#define CC_COUNTRY                                     29
#define CC_GMT_OFFSET                                  30
#define CC_ADDRESS                                     31
#define CC_TAX_PERCENTAGE                              32
#define CC_SCD                                         33
#define CC_NULLS                                       34
#define CALL_CENTER_END                                34
#define CATALOG_PAGE_START                             35
#define CP_CATALOG_PAGE_SK                             35
#define CP_CATALOG_PAGE_ID                             36
#define CP_START_DATE_ID                               37
#define CP_END_DATE_ID                                 38
#define CP_PROMO_ID                                    39
#define CP_DEPARTMENT                                  40
#define CP_CATALOG_NUMBER                              41
#define CP_CATALOG_PAGE_NUMBER                         42
#define CP_DESCRIPTION                                 43
#define CP_TYPE                                        44
#define CP_NULLS                                       45
#define CATALOG_PAGE_END                               45
#define CATALOG_RETURNS_START                          46
#define CR_RETURNED_DATE_SK                            46
#define CR_RETURNED_TIME_SK                            47
#define CR_ITEM_SK                                     48
#define CR_REFUNDED_CUSTOMER_SK                        49
#define CR_REFUNDED_CDEMO_SK                           50
#define CR_REFUNDED_HDEMO_SK                           51
#define CR_REFUNDED_ADDR_SK                            52
#define CR_RETURNING_CUSTOMER_SK                       53
#define CR_RETURNING_CDEMO_SK                          54
#define CR_RETURNING_HDEMO_SK                          55
#define CR_RETURNING_ADDR_SK                           56
#define CR_CALL_CENTER_SK                              57
#define CR_CATALOG_PAGE_SK                             58
#define CR_SHIP_MODE_SK                                59
#define CR_WAREHOUSE_SK                                60
#define CR_REASON_SK                                   61
#define CR_ORDER_NUMBER                                62
#define CR_PRICING_QUANTITY                            63
#define CR_PRICING_NET_PAID                            64
#define CR_PRICING_EXT_TAX                             65
#define CR_PRICING_NET_PAID_INC_TAX                    66
#define CR_PRICING_FEE                                 67
#define CR_PRICING_EXT_SHIP_COST                       68
#define CR_PRICING_REFUNDED_CASH                       69
#define CR_PRICING_REVERSED_CHARGE                     70
#define CR_PRICING_STORE_CREDIT                        71
#define CR_PRICING_NET_LOSS                            72
#define CR_NULLS                                       73
#define CR_PRICING                                     74
#define CATALOG_RETURNS_END                            74
#define CATALOG_SALES_START                            75
#define CS_SOLD_DATE_SK                                75
#define CS_SOLD_TIME_SK                                76
#define CS_SHIP_DATE_SK                                77
#define CS_BILL_CUSTOMER_SK                            78
#define CS_BILL_CDEMO_SK                               79
#define CS_BILL_HDEMO_SK                               80
#define CS_BILL_ADDR_SK                                81
#define CS_SHIP_CUSTOMER_SK                            82
#define CS_SHIP_CDEMO_SK                               83
#define CS_SHIP_HDEMO_SK                               84
#define CS_SHIP_ADDR_SK                                85
#define CS_CALL_CENTER_SK                              86
#define CS_CATALOG_PAGE_SK                             87
#define CS_SHIP_MODE_SK                                88
#define CS_WAREHOUSE_SK                                89
#define CS_SOLD_ITEM_SK                                90
#define CS_PROMO_SK                                    91
#define CS_ORDER_NUMBER                                92
#define CS_PRICING_QUANTITY                            93
#define CS_PRICING_WHOLESALE_COST                      94
#define CS_PRICING_LIST_PRICE                          95
#define CS_PRICING_SALES_PRICE                         96
#define CS_PRICING_COUPON_AMT                          97
#define CS_PRICING_EXT_SALES_PRICE                     98
#define CS_PRICING_EXT_DISCOUNT_AMOUNT                 99
#define CS_PRICING_EXT_WHOLESALE_COST                  100
#define CS_PRICING_EXT_LIST_PRICE                      101
#define CS_PRICING_EXT_TAX                             102
#define CS_PRICING_EXT_SHIP_COST                       103
#define CS_PRICING_NET_PAID                            104
#define CS_PRICING_NET_PAID_INC_TAX                    105
#define CS_PRICING_NET_PAID_INC_SHIP                   106
#define CS_PRICING_NET_PAID_INC_SHIP_TAX               107
#define CS_PRICING_NET_PROFIT                          108
#define CS_PRICING                                     109
#define CS_PERMUTE                                     110
#define CS_NULLS                                       111
#define CR_IS_RETURNED                                 112
#define CS_PERMUTATION                                 113
#define CATALOG_SALES_END                              113
#define CUSTOMER_START                                 114
#define C_CUSTOMER_SK                                  114
#define C_CUSTOMER_ID                                  115
#define C_CURRENT_CDEMO_SK                             116
#define C_CURRENT_HDEMO_SK                             117
#define C_CURRENT_ADDR_SK                              118
#define C_FIRST_SHIPTO_DATE_ID                         119
#define C_FIRST_SALES_DATE_ID                          120
#define C_SALUTATION                                   121
#define C_FIRST_NAME                                   122
#define C_LAST_NAME                                    123
#define C_PREFERRED_CUST_FLAG                          124
#define C_BIRTH_DAY                                    125
#define C_BIRTH_MONTH                                  126
#define C_BIRTH_YEAR                                   127
#define C_BIRTH_COUNTRY                                128
#define C_LOGIN                                        129
#define C_EMAIL_ADDRESS                                130
#define C_LAST_REVIEW_DATE                             131
#define C_NULLS                                        132
#define CUSTOMER_END                                   132
#define CUSTOMER_ADDRESS_START                         133
#define CA_ADDRESS_SK                                  133
#define CA_ADDRESS_ID                                  134
#define CA_ADDRESS_STREET_NUM                          135
#define CA_ADDRESS_STREET_NAME1                        136
#define CA_ADDRESS_STREET_TYPE                         137
#define CA_ADDRESS_SUITE_NUM                           138
#define CA_ADDRESS_CITY                                139
#define CA_ADDRESS_COUNTY                              140
#define CA_ADDRESS_STATE                               141
#define CA_ADDRESS_ZIP                                 142
#define CA_ADDRESS_COUNTRY                             143
#define CA_ADDRESS_GMT_OFFSET                          144
#define CA_LOCATION_TYPE                               145
#define CA_NULLS                                       146
#define CA_ADDRESS                                     147
#define CA_ADDRESS_STREET_NAME2                        148
#define CUSTOMER_ADDRESS_END                           148
#define CUSTOMER_DEMOGRAPHICS_START                    149
#define CD_DEMO_SK                                     149
#define CD_GENDER                                      150
#define CD_MARITAL_STATUS                              151
#define CD_EDUCATION_STATUS                            152
#define CD_PURCHASE_ESTIMATE                           153
#define CD_CREDIT_RATING                               154
#define CD_DEP_COUNT                                   155
#define CD_DEP_EMPLOYED_COUNT                          156
#define CD_DEP_COLLEGE_COUNT                           157
#define CD_NULLS                                       158
#define CUSTOMER_DEMOGRAPHICS_END                      158
#define DATE_START                                     159
#define D_DATE_SK                                      159
#define D_DATE_ID                                      160
#define D_DATE                                         161
#define D_MONTH_SEQ                                    162
#define D_WEEK_SEQ                                     163
#define D_QUARTER_SEQ                                  164
#define D_YEAR                                         165
#define D_DOW                                          166
#define D_MOY                                          167
#define D_DOM                                          168
#define D_QOY                                          169
#define D_FY_YEAR                                      170
#define D_FY_QUARTER_SEQ                               171
#define D_FY_WEEK_SEQ                                  172
#define D_DAY_NAME                                     173
#define D_QUARTER_NAME                                 174
#define D_HOLIDAY                                      175
#define D_WEEKEND                                      176
#define D_FOLLOWING_HOLIDAY                            177
#define D_FIRST_DOM                                    178
#define D_LAST_DOM                                     179
#define D_SAME_DAY_LY                                  180
#define D_SAME_DAY_LQ                                  181
#define D_CURRENT_DAY                                  182
#define D_CURRENT_WEEK                                 183
#define D_CURRENT_MONTH                                184
#define D_CURRENT_QUARTER                              185
#define D_CURRENT_YEAR                                 186
#define D_NULLS                                        187
#define DATE_END                                       187
#define HOUSEHOLD_DEMOGRAPHICS_START                   188
#define HD_DEMO_SK                                     188
#define HD_INCOME_BAND_ID                              189
#define HD_BUY_POTENTIAL                               190
#define HD_DEP_COUNT                                   191
#define HD_VEHICLE_COUNT                               192
#define HD_NULLS                                       193
#define HOUSEHOLD_DEMOGRAPHICS_END                     193
#define INCOME_BAND_START                              194
#define IB_INCOME_BAND_ID                              194
#define IB_LOWER_BOUND                                 195
#define IB_UPPER_BOUND                                 196
#define IB_NULLS                                       197
#define INCOME_BAND_END                                197
#define INVENTORY_START                                198
#define INV_DATE_SK                                    198
#define INV_ITEM_SK                                    199
#define INV_WAREHOUSE_SK                               200
#define INV_QUANTITY_ON_HAND                           201
#define INV_NULLS                                      202
#define INVENTORY_END                                  202
#define ITEM_START                                     203
#define I_ITEM_SK                                      203
#define I_ITEM_ID                                      204
#define I_REC_START_DATE_ID                            205
#define I_REC_END_DATE_ID                              206
#define I_ITEM_DESC                                    207
#define I_CURRENT_PRICE                                208
#define I_WHOLESALE_COST                               209
#define I_BRAND_ID                                     210
#define I_BRAND                                        211
#define I_CLASS_ID                                     212
#define I_CLASS                                        213
#define I_CATEGORY_ID                                  214
#define I_CATEGORY                                     215
#define I_MANUFACT_ID                                  216
#define I_MANUFACT                                     217
#define I_SIZE                                         218
#define I_FORMULATION                                  219
#define I_COLOR                                        220
#define I_UNITS                                        221
#define I_CONTAINER                                    222
#define I_MANAGER_ID                                   223
#define I_PRODUCT_NAME                                 224
#define I_NULLS                                        225
#define I_SCD                                          226
#define I_PROMO_SK                                     227
#define ITEM_END                                       227
#define PROMOTION_START                                228
#define P_PROMO_SK                                     228
#define P_PROMO_ID                                     229
#define P_START_DATE_ID                                230
#define P_END_DATE_ID                                  231
#define P_ITEM_SK                                      232
#define P_COST                                         233
#define P_RESPONSE_TARGET                              234
#define P_PROMO_NAME                                   235
#define P_CHANNEL_DMAIL                                236
#define P_CHANNEL_EMAIL                                237
#define P_CHANNEL_CATALOG                              238
#define P_CHANNEL_TV                                   239
#define P_CHANNEL_RADIO                                240
#define P_CHANNEL_PRESS                                241
#define P_CHANNEL_EVENT                                242
#define P_CHANNEL_DEMO                                 243
#define P_CHANNEL_DETAILS                              244
#define P_PURPOSE                                      245
#define P_DISCOUNT_ACTIVE                              246
#define P_NULLS                                        247
#define PROMOTION_END                                  247
#define REASON_START                                   248
#define R_REASON_SK                                    248
#define R_REASON_ID                                    249
#define R_REASON_DESCRIPTION                           250
#define R_NULLS                                        251
#define REASON_END                                     251
#define SHIP_MODE_START                                252
#define SM_SHIP_MODE_SK                                252
#define SM_SHIP_MODE_ID                                253
#define SM_TYPE                                        254
#define SM_CODE                                        255
#define SM_CONTRACT                                    256
#define SM_CARRIER                                     257
#define SM_NULLS                                       258
#define SHIP_MODE_END                                  258
#define STORE_START                                    259
#define W_STORE_SK                                     259
#define W_STORE_ID                                     260
#define W_STORE_REC_START_DATE_ID                      261
#define W_STORE_REC_END_DATE_ID                        262
#define W_STORE_CLOSED_DATE_ID                         263
#define W_STORE_NAME                                   264
#define W_STORE_EMPLOYEES                              265
#define W_STORE_FLOOR_SPACE                            266
#define W_STORE_HOURS                                  267
#define W_STORE_MANAGER                                268
#define W_STORE_MARKET_ID                              269
#define W_STORE_TAX_PERCENTAGE                         270
#define W_STORE_GEOGRAPHY_CLASS                        271
#define W_STORE_MARKET_DESC                            272
#define W_STORE_MARKET_MANAGER                         273
#define W_STORE_DIVISION_ID                            274
#define W_STORE_DIVISION_NAME                          275
#define W_STORE_COMPANY_ID                             276
#define W_STORE_COMPANY_NAME                           277
#define W_STORE_ADDRESS_STREET_NUM                     278
#define W_STORE_ADDRESS_STREET_NAME1                   279
#define W_STORE_ADDRESS_STREET_TYPE                    280
#define W_STORE_ADDRESS_SUITE_NUM                      281
#define W_STORE_ADDRESS_CITY                           282
#define W_STORE_ADDRESS_COUNTY                         283
#define W_STORE_ADDRESS_STATE                          284
#define W_STORE_ADDRESS_ZIP                            285
#define W_STORE_ADDRESS_COUNTRY                        286
#define W_STORE_ADDRESS_GMT_OFFSET                     287
#define W_STORE_NULLS                                  288
#define W_STORE_TYPE                                   289
#define W_STORE_SCD                                    290
#define W_STORE_ADDRESS                                291
#define STORE_END                                      291
#define STORE_RETURNS_START                            292
#define SR_RETURNED_DATE_SK                            292
#define SR_RETURNED_TIME_SK                            293
#define SR_ITEM_SK                                     294
#define SR_CUSTOMER_SK                                 295
#define SR_CDEMO_SK                                    296
#define SR_HDEMO_SK                                    297
#define SR_ADDR_SK                                     298
#define SR_STORE_SK                                    299
#define SR_REASON_SK                                   300
#define SR_TICKET_NUMBER                               301
#define SR_PRICING_QUANTITY                            302
#define SR_PRICING_NET_PAID                            303
#define SR_PRICING_EXT_TAX                             304
#define SR_PRICING_NET_PAID_INC_TAX                    305
#define SR_PRICING_FEE                                 306
#define SR_PRICING_EXT_SHIP_COST                       307
#define SR_PRICING_REFUNDED_CASH                       308
#define SR_PRICING_REVERSED_CHARGE                     309
#define SR_PRICING_STORE_CREDIT                        310
#define SR_PRICING_NET_LOSS                            311
#define SR_PRICING                                     312
#define SR_NULLS                                       313
#define STORE_RETURNS_END                              313
#define STORE_SALES_START                              314
#define SS_SOLD_DATE_SK                                314
#define SS_SOLD_TIME_SK                                315
#define SS_SOLD_ITEM_SK                                316
#define SS_SOLD_CUSTOMER_SK                            317
#define SS_SOLD_CDEMO_SK                               318
#define SS_SOLD_HDEMO_SK                               319
#define SS_SOLD_ADDR_SK                                320
#define SS_SOLD_STORE_SK                               321
#define SS_SOLD_PROMO_SK                               322
#define SS_TICKET_NUMBER                               323
#define SS_PRICING_QUANTITY                            324
#define SS_PRICING_WHOLESALE_COST                      325
#define SS_PRICING_LIST_PRICE                          326
#define SS_PRICING_SALES_PRICE                         327
#define SS_PRICING_COUPON_AMT                          328
#define SS_PRICING_EXT_SALES_PRICE                     329
#define SS_PRICING_EXT_WHOLESALE_COST                  330
#define SS_PRICING_EXT_LIST_PRICE                      331
#define SS_PRICING_EXT_TAX                             332
#define SS_PRICING_NET_PAID                            333
#define SS_PRICING_NET_PAID_INC_TAX                    334
#define SS_PRICING_NET_PROFIT                          335
#define SR_IS_RETURNED                                 336
#define SS_PRICING                                     337
#define SS_NULLS                                       338
#define SS_PERMUTATION                                 339
#define STORE_SALES_END                                339
#define TIME_START                                     340
#define T_TIME_SK                                      340
#define T_TIME_ID                                      341
#define T_TIME                                         342
#define T_HOUR                                         343
#define T_MINUTE                                       344
#define T_SECOND                                       345
#define T_AM_PM                                        346
#define T_SHIFT                                        347
#define T_SUB_SHIFT                                    348
#define T_MEAL_TIME                                    349
#define T_NULLS                                        350
#define TIME_END                                       350
#define WAREHOUSE_START                                351
#define W_WAREHOUSE_SK                                 351
#define W_WAREHOUSE_ID                                 352
#define W_WAREHOUSE_NAME                               353
#define W_WAREHOUSE_SQ_FT                              354
#define W_ADDRESS_STREET_NUM                           355
#define W_ADDRESS_STREET_NAME1                         356
#define W_ADDRESS_STREET_TYPE                          357
#define W_ADDRESS_SUITE_NUM                            358
#define W_ADDRESS_CITY                                 359
#define W_ADDRESS_COUNTY                               360
#define W_ADDRESS_STATE                                361
#define W_ADDRESS_ZIP                                  362
#define W_ADDRESS_COUNTRY                              363
#define W_ADDRESS_GMT_OFFSET                           364
#define W_NULLS                                        365
#define W_WAREHOUSE_ADDRESS                            366
#define WAREHOUSE_END                                  366
#define WEB_PAGE_START                                 367
#define WP_PAGE_SK                                     367
#define WP_PAGE_ID                                     368
#define WP_REC_START_DATE_ID                           369
#define WP_REC_END_DATE_ID                             370
#define WP_CREATION_DATE_SK                            371
#define WP_ACCESS_DATE_SK                              372
#define WP_AUTOGEN_FLAG                                373
#define WP_CUSTOMER_SK                                 374
#define WP_URL                                         375
#define WP_TYPE                                        376
#define WP_CHAR_COUNT                                  377
#define WP_LINK_COUNT                                  378
#define WP_IMAGE_COUNT                                 379
#define WP_MAX_AD_COUNT                                380
#define WP_NULLS                                       381
#define WP_SCD                                         382
#define WEB_PAGE_END                                   382
#define WEB_RETURNS_START                              383
#define WR_RETURNED_DATE_SK                            383
#define WR_RETURNED_TIME_SK                            384
#define WR_ITEM_SK                                     385
#define WR_REFUNDED_CUSTOMER_SK                        386
#define WR_REFUNDED_CDEMO_SK                           387
#define WR_REFUNDED_HDEMO_SK                           388
#define WR_REFUNDED_ADDR_SK                            389
#define WR_RETURNING_CUSTOMER_SK                       390
#define WR_RETURNING_CDEMO_SK                          391
#define WR_RETURNING_HDEMO_SK                          392
#define WR_RETURNING_ADDR_SK                           393
#define WR_WEB_PAGE_SK                                 394
#define WR_REASON_SK                                   395
#define WR_ORDER_NUMBER                                396
#define WR_PRICING_QUANTITY                            397
#define WR_PRICING_NET_PAID                            398
#define WR_PRICING_EXT_TAX                             399
#define WR_PRICING_NET_PAID_INC_TAX                    400
#define WR_PRICING_FEE                                 401
#define WR_PRICING_EXT_SHIP_COST                       402
#define WR_PRICING_REFUNDED_CASH                       403
#define WR_PRICING_REVERSED_CHARGE                     404
#define WR_PRICING_STORE_CREDIT                        405
#define WR_PRICING_NET_LOSS                            406
#define WR_PRICING                                     407
#define WR_NULLS                                       408
#define WEB_RETURNS_END                                408
#define WEB_SALES_START                                409
#define WS_SOLD_DATE_SK                                409
#define WS_SOLD_TIME_SK                                410
#define WS_SHIP_DATE_SK                                411
#define WS_ITEM_SK                                     412
#define WS_BILL_CUSTOMER_SK                            413
#define WS_BILL_CDEMO_SK                               414
#define WS_BILL_HDEMO_SK                               415
#define WS_BILL_ADDR_SK                                416
#define WS_SHIP_CUSTOMER_SK                            417
#define WS_SHIP_CDEMO_SK                               418
#define WS_SHIP_HDEMO_SK                               419
#define WS_SHIP_ADDR_SK                                420
#define WS_WEB_PAGE_SK                                 421
#define WS_WEB_SITE_SK                                 422
#define WS_SHIP_MODE_SK                                423
#define WS_WAREHOUSE_SK                                424
#define WS_PROMO_SK                                    425
#define WS_ORDER_NUMBER                                426
#define WS_PRICING_QUANTITY                            427
#define WS_PRICING_WHOLESALE_COST                      428
#define WS_PRICING_LIST_PRICE                          429
#define WS_PRICING_SALES_PRICE                         430
#define WS_PRICING_EXT_DISCOUNT_AMT                    431
#define WS_PRICING_EXT_SALES_PRICE                     432
#define WS_PRICING_EXT_WHOLESALE_COST                  433
#define WS_PRICING_EXT_LIST_PRICE                      434
#define WS_PRICING_EXT_TAX                             435
#define WS_PRICING_COUPON_AMT                          436
#define WS_PRICING_EXT_SHIP_COST                       437
#define WS_PRICING_NET_PAID                            438
#define WS_PRICING_NET_PAID_INC_TAX                    439
#define WS_PRICING_NET_PAID_INC_SHIP                   440
#define WS_PRICING_NET_PAID_INC_SHIP_TAX               441
#define WS_PRICING_NET_PROFIT                          442
#define WS_PRICING                                     443
#define WS_NULLS                                       444
#define WR_IS_RETURNED                                 445
#define WS_PERMUTATION                                 446
#define WEB_SALES_END                                  446
#define WEB_SITE_START                                 447
#define WEB_SITE_SK                                    447
#define WEB_SITE_ID                                    448
#define WEB_REC_START_DATE_ID                          449
#define WEB_REC_END_DATE_ID                            450
#define WEB_NAME                                       451
#define WEB_OPEN_DATE                                  452
#define WEB_CLOSE_DATE                                 453
#define WEB_CLASS                                      454
#define WEB_MANAGER                                    455
#define WEB_MARKET_ID                                  456
#define WEB_MARKET_CLASS                               457
#define WEB_MARKET_DESC                                458
#define WEB_MARKET_MANAGER                             459
#define WEB_COMPANY_ID                                 460
#define WEB_COMPANY_NAME                               461
#define WEB_ADDRESS_STREET_NUM                         462
#define WEB_ADDRESS_STREET_NAME1                       463
#define WEB_ADDRESS_STREET_TYPE                        464
#define WEB_ADDRESS_SUITE_NUM                          465
#define WEB_ADDRESS_CITY                               466
#define WEB_ADDRESS_COUNTY                             467
#define WEB_ADDRESS_STATE                              468
#define WEB_ADDRESS_ZIP                                469
#define WEB_ADDRESS_COUNTRY                            470
#define WEB_ADDRESS_GMT_OFFSET                         471
#define WEB_TAX_PERCENTAGE                             472
#define WEB_NULLS                                      473
#define WEB_ADDRESS                                    474
#define WEB_SCD                                        475
#define WEB_SITE_END                                   475
#define DBGEN_VERSION_START                            476
#define DV_VERSION                                     476
#define DV_CREATE_DATE                                 477
#define DV_CREATE_TIME                                 478
#define DV_CMDLINE_ARGS                                479
#define VALIDATE_STREAM                                480
#define DBGEN_VERSION_END                              480
#define S_BRAND_START                                  481
#define S_BRAND_ID                                     481
#define S_BRAND_SUBCLASS_ID                            482
#define S_BRAND_MANAGER_ID                             483
#define S_BRAND_MANUFACTURER_ID                        484
#define S_BRAND_NAME                                   485
#define S_BRAND_END                                    485
#define S_CUSTOMER_ADDRESS_START                       486
#define S_CADR_ID                                      486
#define S_CADR_ADDRESS_STREET_NUMBER                   487
#define S_CADR_ADDRESS_STREET_NAME1                    488
#define S_CADR_ADDRESS_STREET_NAME2                    489
#define S_CADR_ADDRESS_STREET_TYPE                     490
#define S_CADR_ADDRESS_SUITE_NUM                       491
#define S_CADR_ADDRESS_CITY                            492
#define S_CADR_ADDRESS_COUNTY                          493
#define S_CADR_ADDRESS_STATE                           494
#define S_CADR_ADDRESS_ZIP                             495
#define S_CADR_ADDRESS_COUNTRY                         496
#define S_BADDR_ADDRESS                                497
#define S_CUSTOMER_ADDRESS_END                         497
#define S_CALL_CENTER_START                            498
#define S_CALL_CENTER_ID                               498
#define S_CALL_CENTER_DIVISION_ID                      499
#define S_CALL_CENTER_OPEN_DATE                        500
#define S_CALL_CENTER_CLOSED_DATE                      501
#define S_CALL_CENTER_NAME                             502
#define S_CALL_CENTER_CLASS                            503
#define S_CALL_CENTER_EMPLOYEES                        504
#define S_CALL_CENTER_SQFT                             505
#define S_CALL_CENTER_HOURS                            506
#define S_CALL_CENTER_MANAGER_ID                       507
#define S_CALL_CENTER_MARKET_ID                        508
#define S_CALL_CENTER_ADDRESS_ID                       509
#define S_CALL_CENTER_TAX_PERCENTAGE                   510
#define S_CALL_CENTER_SCD                              511
#define S_CALL_CENTER_END                              511
#define S_CATALOG_START                                512
#define S_CATALOG_NUMBER                               512
#define S_CATALOG_START_DATE                           513
#define S_CATALOG_END_DATE                             514
#define S_CATALOG_DESC                                 515
#define S_CATALOG_TYPE                                 516
#define S_CATALOG_END                                  516
#define S_CATALOG_ORDER_START                          517
#define S_CORD_ID                                      517
#define S_CORD_BILL_CUSTOMER_ID                        518
#define S_CORD_SHIP_CUSTOMER_ID                        519
#define S_CORD_ORDER_DATE                              520
#define S_CORD_ORDER_TIME                              521
#define S_CORD_SHIP_MODE_ID                            522
#define S_CORD_CALL_CENTER_ID                          523
#define S_CLIN_ITEM_ID                                 524
#define S_CORD_COMMENT                                 525
#define S_CATALOG_ORDER_END                            525
#define S_CATALOG_ORDER_LINEITEM_START                 526
#define S_CLIN_ORDER_ID                                526
#define S_CLIN_LINE_NUMBER                             527
#define S_CLIN_PROMOTION_ID                            528
#define S_CLIN_QUANTITY                                529
#define S_CLIN_COUPON_AMT                              530
#define S_CLIN_WAREHOUSE_ID                            531
#define S_CLIN_SHIP_DATE                               532
#define S_CLIN_CATALOG_ID                              533
#define S_CLIN_CATALOG_PAGE_ID                         534
#define S_CLIN_PRICING                                 535
#define S_CLIN_SHIP_COST                               536
#define S_CLIN_IS_RETURNED                             537
#define S_CLIN_PERMUTE                                 538
#define S_CATALOG_ORDER_LINEITEM_END                   538
#define S_CATALOG_PAGE_START                           539
#define S_CATALOG_PAGE_CATALOG_NUMBER                  539
#define S_CATALOG_PAGE_NUMBER                          540
#define S_CATALOG_PAGE_DEPARTMENT                      541
#define S_CP_ID                                        542
#define S_CP_START_DATE                                543
#define S_CP_END_DATE                                  544
#define S_CP_DESCRIPTION                               545
#define S_CP_TYPE                                      546
#define S_CATALOG_PAGE_END                             546
#define S_CATALOG_PROMOTIONAL_ITEM_START               547
#define S_CATALOG_PROMOTIONAL_ITEM_CATALOG_NUMBER      547
#define S_CATALOG_PROMOTIONAL_ITEM_CATALOG_PAGE_NUMBER 548
#define S_CATALOG_PROMOTIONAL_ITEM_ITEM_ID             549
#define S_CATALOG_PROMOTIONAL_ITEM_PROMOTION_ID        550
#define S_CATALOG_PROMOTIONAL_ITEM_END                 550
#define S_CATALOG_RETURNS_START                        551
#define S_CRET_CALL_CENTER_ID                          551
#define S_CRET_ORDER_ID                                552
#define S_CRET_LINE_NUMBER                             553
#define S_CRET_ITEM_ID                                 554
#define S_CRET_RETURN_CUSTOMER_ID                      555
#define S_CRET_REFUND_CUSTOMER_ID                      556
#define S_CRET_DATE                                    557
#define S_CRET_TIME                                    558
#define S_CRET_QUANTITY                                559
#define S_CRET_AMOUNT                                  560
#define S_CRET_TAX                                     561
#define S_CRET_FEE                                     562
#define S_CRET_SHIP_COST                               563
#define S_CRET_REFUNDED_CASH                           564
#define S_CRET_REVERSED_CHARGE                         565
#define S_CRET_MERCHANT_CREDIT                         566
#define S_CRET_REASON_ID                               567
#define S_CRET_PRICING                                 568
#define S_CRET_SHIPMODE_ID                             569
#define S_CRET_WAREHOUSE_ID                            570
#define S_CRET_CATALOG_PAGE_ID                         571
#define S_CATALOG_RETURNS_END                          571
#define S_CATEGORY_START                               572
#define S_CATEGORY_ID                                  572
#define S_CATEGORY_NAME                                573
#define S_CATEGORY_DESC                                574
#define S_CATEGORY_END                                 574
#define S_CLASS_START                                  575
#define S_CLASS_ID                                     575
#define S_CLASS_SUBCAT_ID                              576
#define S_CLASS_DESC                                   577
#define S_CLASS_END                                    577
#define S_COMPANY_START                                578
#define S_COMPANY_ID                                   578
#define S_COMPANY_NAME                                 579
#define S_COMPANY_END                                  579
#define S_CUSTOMER_START                               580
#define S_CUST_ID                                      580
#define S_CUST_SALUTATION                              581
#define S_CUST_LAST_NAME                               582
#define S_CUST_FIRST_NAME                              583
#define S_CUST_PREFERRED_FLAG                          584
#define S_CUST_BIRTH_DATE                              585
#define S_CUST_FIRST_PURCHASE_DATE                     586
#define S_CUST_FIRST_SHIPTO_DATE                       587
#define S_CUST_BIRTH_COUNTRY                           588
#define S_CUST_LOGIN                                   589
#define S_CUST_EMAIL                                   590
#define S_CUST_LAST_LOGIN                              591
#define S_CUST_LAST_REVIEW                             592
#define S_CUST_PRIMARY_MACHINE                         593
#define S_CUST_SECONDARY_MACHINE                       594
#define S_CUST_ADDRESS                                 595
#define S_CUST_ADDRESS_STREET_NUM                      596
#define S_CUST_ADDRESS_STREET_NAME1                    597
#define S_CUST_ADDRESS_STREET_NAME2                    598
#define S_CUST_ADDRESS_STREET_TYPE                     599
#define S_CUST_ADDRESS_SUITE_NUM                       600
#define S_CUST_ADDRESS_CITY                            601
#define S_CUST_ADDRESS_ZIP                             602
#define S_CUST_ADDRESS_COUNTY                          603
#define S_CUST_ADDRESS_STATE                           604
#define S_CUST_ADDRESS_COUNTRY                         605
#define S_CUST_LOCATION_TYPE                           606
#define S_CUST_GENDER                                  607
#define S_CUST_MARITAL_STATUS                          608
#define S_CUST_EDUCATION                               609
#define S_CUST_CREDIT_RATING                           610
#define S_CUST_PURCHASE_ESTIMATE                       611
#define S_CUST_BUY_POTENTIAL                           612
#define S_CUST_DEPENDENT_CNT                           613
#define S_CUST_EMPLOYED_CNT                            614
#define S_CUST_COLLEGE_CNT                             615
#define S_CUST_VEHICLE_CNT                             616
#define S_CUST_INCOME                                  617
#define S_CUSTOMER_END                                 617
#define S_DIVISION_START                               618
#define S_DIVISION_ID                                  618
#define S_DIVISION_COMPANY                             619
#define S_DIVISION_NAME                                620
#define S_DIVISION_END                                 620
#define S_INVENTORY_START                              621
#define S_INVN_WAREHOUSE                               621
#define S_INVN_ITEM                                    622
#define S_INVN_DATE                                    623
#define S_INVN_QUANTITY                                624
#define S_INVENTORY_END                                624
#define S_ITEM_START                                   625
#define S_ITEM_ID                                      625
#define S_ITEM_PERMUTE                                 626
#define S_ITEM_PRODUCT_ID                              627
#define S_ITEM_DESC                                    628
#define S_ITEM_LIST_PRICE                              629
#define S_ITEM_WHOLESALE_COST                          630
#define S_ITEM_MANAGER_ID                              631
#define S_ITEM_SIZE                                    632
#define S_ITEM_FORMULATION                             633
#define S_ITEM_FLAVOR                                  634
#define S_ITEM_UNITS                                   635
#define S_ITEM_CONTAINER                               636
#define S_ITEM_SCD                                     637
#define S_ITEM_END                                     637
#define S_MANAGER_START                                638
#define S_MANAGER_ID                                   638
#define S_MANAGER_NAME                                 639
#define S_MANAGER_END                                  639
#define S_MANUFACTURER_START                           640
#define S_MANUFACTURER_ID                              640
#define S_MANUFACTURER_NAME                            641
#define S_MANUFACTURER_END                             641
#define S_MARKET_START                                 642
#define S_MARKET_ID                                    642
#define S_MARKET_CLASS_NAME                            643
#define S_MARKET_DESC                                  644
#define S_MARKET_MANAGER_ID                            645
#define S_MARKET_END                                   645
#define S_PRODUCT_START                                646
#define S_PRODUCT_ID                                   646
#define S_PRODUCT_BRAND_ID                             647
#define S_PRODUCT_NAME                                 648
#define S_PRODUCT_TYPE                                 649
#define S_PRODUCT_END                                  649
#define S_PROMOTION_START                              650
#define S_PROMOTION_ID                                 650
#define S_PROMOTION_ITEM_ID                            651
#define S_PROMOTION_START_DATE                         652
#define S_PROMOTION_END_DATE                           653
#define S_PROMOTION_COST                               654
#define S_PROMOTION_RESPONSE_TARGET                    655
#define S_PROMOTION_DMAIL                              656
#define S_PROMOTION_EMAIL                              657
#define S_PROMOTION_CATALOG                            658
#define S_PROMOTION_TV                                 659
#define S_PROMOTION_RADIO                              660
#define S_PROMOTION_PRESS                              661
#define S_PROMOTION_EVENT                              662
#define S_PROMOTION_DEMO                               663
#define S_PROMOTION_DETAILS                            664
#define S_PROMOTION_PURPOSE                            665
#define S_PROMOTION_DISCOUNT_ACTIVE                    666
#define S_PROMOTION_DISCOUNT_PCT                       667
#define S_PROMOTION_NAME                               668
#define S_PROMOTION_BITFIELD                           669
#define S_PROMOTION_END                                669
#define S_PURCHASE_START                               670
#define S_PURCHASE_ID                                  670
#define S_PURCHASE_STORE_ID                            671
#define S_PURCHASE_CUSTOMER_ID                         672
#define S_PURCHASE_DATE                                673
#define S_PURCHASE_TIME                                674
#define S_PURCHASE_REGISTER                            675
#define S_PURCHASE_CLERK                               676
#define S_PURCHASE_COMMENT                             677
#define S_PURCHASE_PRICING                             678
#define S_PLINE_ITEM_ID                                679
#define S_PURCHASE_END                                 679
#define S_PURCHASE_LINEITEM_START                      680
#define S_PLINE_PURCHASE_ID                            680
#define S_PLINE_NUMBER                                 681
#define S_PLINE_PROMOTION_ID                           682
#define S_PLINE_SALE_PRICE                             683
#define S_PLINE_QUANTITY                               684
#define S_PLINE_COUPON_AMT                             685
#define S_PLINE_COMMENT                                686
#define S_PLINE_PRICING                                687
#define S_PLINE_IS_RETURNED                            688
#define S_PLINE_PERMUTE                                689
#define S_PURCHASE_LINEITEM_END                        689
#define S_REASON_START                                 690
#define S_REASON_ID                                    690
#define S_REASON_DESC                                  691
#define S_REASON_END                                   691
#define S_STORE_START                                  692
#define S_STORE_ID                                     692
#define S_STORE_ADDRESS_ID                             693
#define S_STORE_DIVISION_ID                            694
#define S_STORE_OPEN_DATE                              695
#define S_STORE_CLOSE_DATE                             696
#define S_STORE_NAME                                   697
#define S_STORE_CLASS                                  698
#define S_STORE_EMPLOYEES                              699
#define S_STORE_FLOOR_SPACE                            700
#define S_STORE_HOURS                                  701
#define S_STORE_MARKET_MANAGER_ID                      702
#define S_STORE_MANAGER_ID                             703
#define S_STORE_MARKET_ID                              704
#define S_STORE_GEOGRAPHY_CLASS                        705
#define S_STORE_TAX_PERCENTAGE                         706
#define S_STORE_END                                    706
#define S_STORE_PROMOTIONAL_ITEM_START                 707
#define S_SITM_PROMOTION_ID                            707
#define S_SITM_ITEM_ID                                 708
#define S_SITM_STORE_ID                                709
#define S_STORE_PROMOTIONAL_ITEM_END                   709
#define S_STORE_RETURNS_START                          710
#define S_SRET_STORE_ID                                710
#define S_SRET_PURCHASE_ID                             711
#define S_SRET_LINENUMBER                              712
#define S_SRET_ITEM_ID                                 713
#define S_SRET_CUSTOMER_ID                             714
#define S_SRET_RETURN_DATE                             715
#define S_SRET_RETURN_TIME                             716
#define S_SRET_TICKET_NUMBER                           717
#define S_SRET_RETURN_QUANTITY                         718
#define S_SRET_RETURN_AMT                              719
#define S_SRET_RETURN_TAX                              720
#define S_SRET_RETURN_FEE                              721
#define S_SRET_RETURN_SHIP_COST                        722
#define S_SRET_REFUNDED_CASH                           723
#define S_SRET_REVERSED_CHARGE                         724
#define S_SRET_MERCHANT_CREDIT                         725
#define S_SRET_REASON_ID                               726
#define S_SRET_PRICING                                 727
#define S_STORE_RETURNS_END                            727
#define S_SUBCATEGORY_START                            728
#define S_SBCT_ID                                      728
#define S_SBCT_CATEGORY_ID                             729
#define S_SBCT_NAME                                    730
#define S_SBCT_DESC                                    731
#define S_SUBCATEGORY_END                              731
#define S_SUBCLASS_START                               732
#define S_SUBC_ID                                      732
#define S_SUBC_CLASS_ID                                733
#define S_SUBC_NAME                                    734
#define S_SUBC_DESC                                    735
#define S_SUBCLASS_END                                 735
#define S_WAREHOUSE_START                              736
#define S_WRHS_ID                                      736
#define S_WRHS_DESC                                    737
#define S_WRHS_SQFT                                    738
#define S_WRHS_ADDRESS_ID                              739
#define S_WAREHOUSE_END                                739
#define S_WEB_ORDER_START                              740
#define S_WORD_ID                                      740
#define S_WORD_BILL_CUSTOMER_ID                        741
#define S_WORD_SHIP_CUSTOMER_ID                        742
#define S_WORD_ORDER_DATE                              743
#define S_WORD_ORDER_TIME                              744
#define S_WORD_SHIP_MODE_ID                            745
#define S_WORD_WEB_SITE_ID                             746
#define S_WORD_COMMENT                                 747
#define S_WLIN_ITEM_ID                                 748
#define S_WEB_ORDER_END                                748
#define S_WEB_ORDER_LINEITEM_START                     749
#define S_WLIN_ID                                      749
#define S_WLIN_LINE_NUMBER                             750
#define S_WLIN_PROMOTION_ID                            751
#define S_WLIN_QUANTITY                                752
#define S_WLIN_COUPON_AMT                              753
#define S_WLIN_WAREHOUSE_ID                            754
#define S_WLIN_SHIP_DATE                               755
#define S_WLIN_WEB_PAGE_ID                             756
#define S_WLIN_PRICING                                 757
#define S_WLIN_SHIP_COST                               758
#define S_WLIN_IS_RETURNED                             759
#define S_WLIN_PERMUTE                                 760
#define S_WEB_ORDER_LINEITEM_END                       760
#define S_WEB_PAGE_START                               761
#define S_WPAG_SITE_ID                                 761
#define S_WPAG_ID                                      762
#define S_WPAG_CREATE_DATE                             763
#define S_WPAG_ACCESS_DATE                             764
#define S_WPAG_AUTOGEN_FLAG                            765
#define S_WPAG_DEPARTMENT                              766
#define S_WPAG_URL                                     767
#define S_WPAG_TYPE                                    768
#define S_WPAG_CHAR_CNT                                769
#define S_WPAG_LINK_CNT                                770
#define S_WPAG_IMAGE_CNT                               771
#define S_WPAG_MAX_AD_CNT                              772
#define S_WPAG_PERMUTE                                 773
#define S_WEB_PAGE_END                                 773
#define S_WEB_PROMOTIONAL_ITEM_START                   774
#define S_WITM_SITE_ID                                 774
#define S_WITM_PAGE_ID                                 775
#define S_WITM_ITEM_ID                                 776
#define S_WITM_PROMOTION_ID                            777
#define S_WEB_PROMOTIONAL_ITEM_END                     777
#define S_WEB_RETURNS_START                            778
#define S_WRET_SITE_ID                                 778
#define S_WRET_ORDER_ID                                779
#define S_WRET_LINE_NUMBER                             780
#define S_WRET_ITEM_ID                                 781
#define S_WRET_RETURN_CUST_ID                          782
#define S_WRET_REFUND_CUST_ID                          783
#define S_WRET_RETURN_DATE                             784
#define S_WRET_RETURN_TIME                             785
#define S_WRET_REASON_ID                               786
#define S_WRET_PRICING                                 787
#define S_WEB_RETURNS_END                              787
#define S_WEB_SITE_START                               788
#define S_WSIT_ID                                      788
#define S_WSIT_OPEN_DATE                               789
#define S_WSIT_CLOSE_DATE                              790
#define S_WSIT_NAME                                    791
#define S_WSIT_ADDRESS_ID                              792
#define S_WSIT_DIVISION_ID                             793
#define S_WSIT_CLASS                                   794
#define S_WSIT_MANAGER_ID                              795
#define S_WSIT_MARKET_ID                               796
#define S_WSIT_TAX_PERCENTAGE                          797
#define S_WEB_SITE_END                                 797
#define S_ZIPG_START                                   798
#define S_ZIPG_ZIP                                     798
#define S_ZIPG_GMT                                     799
#define S_ZIPG_END                                     799
#define MAX_COLUMN                                     799
#endif
