create table customer(
	c_customer_sk integer not null,
	c_customer_id varchar(16) not null,
	c_current_cdemo_sk integer,
	c_current_hdemo_sk integer,
	c_current_addr_sk integer,
	c_first_shipto_date_sk integer,
	c_first_sales_date_sk integer,
	c_salutation varchar(10),
	c_first_name varchar(20),
	c_last_name varchar(30),
	c_preferred_cust_flag varchar(1),
	c_birth_day integer,
	c_birth_month integer,
	c_birth_year integer,
	c_birth_country varchar(20),
	c_login varchar(13),
	c_email_address varchar(50),
	c_last_review_date_sk integer
);
